#include "canonicalEncoder.hpp"
#include "tableDecoder.hpp"
#include <algorithm>
#include <sdsl/bitmagic.hpp>
#include "constants.hpp"

struct comparator {
	bool operator() (std::pair<unsigned long, unsigned long> i,std::pair<unsigned long, unsigned long> j) {
		return (i.second>j.second);
	}
};

canonicalEncoder::canonicalEncoder() {
	decode = new tableDecoder();
}

void canonicalEncoder::include(unsigned long bitpattern) {
	patternCount[bitpattern]++;
}

bool canonicalEncoder::exclude(unsigned long bitpattern) { return false;} //not yet implemented

void canonicalEncoder::calcCode() {
	//convert hashmap to vector for sorting
	unsigned long maxLength = 0;
	std::vector<std::pair<unsigned long, unsigned long> > patternCountVector;
	patternCountVector.resize(patternCount.size());
	std::vector<std::pair<unsigned long, unsigned long> >::iterator vectIt = patternCountVector.begin();
	std::map<unsigned long, unsigned long>::iterator mapIt;
	for(mapIt = patternCount.begin(); mapIt != patternCount.end(); mapIt++) {
		unsigned long length = sdsl::bit_magic::l1BP(mapIt->first) + 1;
		maxLength = std::max(maxLength, length);
		#ifdef __DEBUG__
		if (maxLength > 64) debug = true;
		if (debug) {
			printf("canonicalEncoder.calcCode(): length = %ld, maxlength = %ld, mapIt->first = %llx\n", length, maxLength, mapIt->first);
		}
		debug = false;
		#endif //debug
		*vectIt = *mapIt;
		vectIt++;
	}
	//sort
	std::sort(patternCountVector.begin(), patternCountVector.end(), comparator());
	decode->init(patternCountVector.size() + 1, maxLength);		//+ 1 is for the zero length encoding we do not want to use
	//convert vector back to hash and fill decoder
	predictedSize = 0;
	for(unsigned long i = 0; i < patternCountVector.size(); i++){
		unsigned long codeWithLength = i + 2;
		patternCount[patternCountVector[i].first] = codeWithLength;
		unsigned long length = sdsl::bit_magic::l1BP(codeWithLength);
		unsigned long codeWithoutLength = codeWithLength - (1 << length);
		predictedSize += length * patternCountVector[i].second;
		decode->add(codeWithoutLength, length, patternCountVector[i].first);
	}
}

unsigned long canonicalEncoder::predictSize() {
	return predictedSize;
}

unsigned long canonicalEncoder::encode(unsigned long bitpattern) {
	unsigned long ret = patternCount[bitpattern];
	length = sdsl::bit_magic::l1BP(ret);
	ret = ret - (1 << length);
	return ret;
}

unsigned long canonicalEncoder::getLength() {
	return length;
}

decoder* canonicalEncoder::getDecoder() {
	return decode;
}
