#ifndef __ENCODER_HPP_INCLUDED__
#define __ENCODER_HPP_INCLUDED__
#include "decoder.hpp"
class encoder {
	public:
		/** include the bitpattern so it will be encoded later. Multiple inclution is necessary for entropy based optimisation
		*
		* @param bitpattern: the pattern to be encoded
		*/
		virtual void include(unsigned long bitpattern) = 0;

		/** undo inclusion. NOT YET IMPLEMENTED

		*/
		virtual bool exclude(unsigned long bitpattern) = 0;

		/** calculate the encoding. all previously included bitpatterns will be encodabte afterwards.
		* @precondition all relevant patterns should be included before.
		* @warning only use once!
		*/
		virtual void calcCode() = 0;
		
		/** predicts the total size (in bits) of all included, encoded bitpatterns
		* @precondition all patterns have been included as often as they are used and calcCode has to be called
		* @return the total size in bits
		*/
		virtual unsigned long predictSize() = 0;

		/** get the code for the given bitpattern
		* @param bitpattern: the bitpattern to be encoded
		* @precondition the given bitpattern has to be included at least once and calcCode has to be called (in that order)
		* @return the code used for the given bitpattern
		* @warning get the length of the code by calling getLength afterwards
		*/
		virtual unsigned long encode(unsigned long bitpattern) = 0;

		/** get the length of the latest accessed code
		* @precondition the code for the bitpattern of interest has to be accessed first via encode
		* @return the length of the latest code
		*/
		virtual unsigned long getLength() = 0;

		/** get a decoder prepared to decode the code
		* @warning there is only one decoder object allocated during constructor call. it will survive the encoders destruction and needs to be destructed by the user.
		* @return a pointer to the single decoder object associated with this encoder. 
		*/
		virtual decoder* getDecoder() = 0;
};
#endif
