bool debug;

#include <stdio.h>
#include <fstream>
#include <time.h>

#include "constants.hpp"

#include <sdsl/vectors.hpp>
#include <sdsl/bitmagic.hpp> //testing only?
#include <sdsl/select_support_mcl.hpp>
#include "escapeRanker.hpp"
using namespace sdsl;
using namespace std;
void output_bitsize(unsigned long x) {
	unsigned long b, B, KiB, MiB, GiB;
	b = x%8;
	x = x/8;
	B = x%1024;
	x = x/1024;
	KiB = x%1024;
	x = x/1024;
	MiB = x%1024;
	x = x/1024;
	GiB = x;
	printf("%dGiB %dMiB %dKiB %dB %db", GiB, MiB, KiB, B, b);
}

int main( int argc, const char* argv[] )
{
debug = false;
	int_vector<1> bv;
	ifstream infile;
	const char* filename = argv[1];
	printf(filename);
	printf("\n");
	char c;
	unsigned int countOnes = 0;
	std::vector<bool> boolvect;

	infile.open(filename);
/*/	for (int i = 0; i < 2000; i++) {
		printf("%c", infile.get());
	}
	return 0;*/
	//determine encoding:
	char zero = infile.get();
	char one = 0;
	while(!one) {
		c = infile.get();
		if (zero != c) one = c;
	}
	infile.seekg (0, ios::beg);

	while(infile.good()) {
		c = infile.get();
		if(c == zero) boolvect.push_back(false);
		else if(c == one) {
			boolvect.push_back(true);
			countOnes++;
		}
	}
	bv.resize(boolvect.size());
	for (unsigned long i = 0; i < boolvect.size();i++) {
		bv[i]= boolvect[i]?1:0;
	}/**/
	printf("original size in bits: ");
	output_bitsize(boolvect.size());
	printf("\n");

	std::clock_t start, end;
	sdsl::stop_watch* stoppie = new sdsl::stop_watch();
#ifdef __CHECK_RANK__
	{
		escapeRanker ibv(bv);
		printf("\nchecking rank and select correctness:\n");fflush(stdout);
		unsigned long rnk = 0, sel = 0;
		for (unsigned long i = 0; i < boolvect.size();i++) {
			if (ibv.rank(i) != rnk) {
				debug = true;
				printf("rankcheck: i/LEVELB = %d, lvlBblock = %lx\n",i/LEVELB, bv.get_int((i/LEVELB) * LEVELB, LEVELB));
				printf("Error: i = %ld, rank = %ld, rnk = %ld\n", i, ibv.rank(i), rnk);
				usleep(200*1000);
			}
			if (bv[i] == 1) {
				rnk++;
				sel = i;
			}
//			if (i > 249999780) {debug = true;printf("%d %d %d\n", i, rnk, sel); fflush(stdout);}
			if (ibv.select(rnk) != sel) {

				debug = true;
				printf("selectcheck: i/LEVELB = %d, lvlBblock = %lx\n",i/LEVELB, bv.get_int((i/LEVELB) * LEVELB, LEVELB));
				printf("Error: i = %ld, select = %ld, sel = %ld\n", i, ibv.select(rnk), sel);
				usleep(200*1000);
			}
		}
	}
#endif //checkRank

#ifdef __MEASURE_RANK__
	{
		escapeRanker ibv(bv);
		unsigned int POSITIONS = 10000000;
		unsigned int REPEATS = 10;
		unsigned int * const position = (unsigned int *)calloc( POSITIONS, sizeof *position );


		for( long long i = 0; i < POSITIONS; i++ ) position[ i ] = rand() % boolvect.size();
		unsigned int dummy = 0x12345678;
		sdsl::stop_watch* stoppie = new sdsl::stop_watch();
		stoppie->start();
		for( int k = REPEATS; k-- != 0; )
			for( int i = 0; i < POSITIONS; i += 10 ) {
				dummy ^= ibv.rank( position[ i ] );
				dummy ^= ibv.rank( position[ i + 1 ] );
				dummy ^= ibv.rank( position[ i + 2 ] );
				dummy ^= ibv.rank( position[ i + 3 ] );
				dummy ^= ibv.rank( position[ i + 4 ] );
				dummy ^= ibv.rank( position[ i + 5 ] );
				dummy ^= ibv.rank( position[ i + 6 ] );
				dummy ^= ibv.rank( position[ i + 7 ] );
				dummy ^= ibv.rank( position[ i + 8 ] );
				dummy ^= ibv.rank( position[ i + 9 ] );
			}
		stoppie->stop();
		end = (long)std::clock();
		printf("\n%d\t%d\n", start, end);
		double s = stoppie->get_user_time() / 1000;
		printf( "%fs, %f ranks/s, hrg%fns/rank, r3j%f \n", s, (double)(REPEATS * POSITIONS) / s, ((double)(1E9 * s)) / (double)(REPEATS * POSITIONS), (double)ibv.size() / boolvect.size() );fflush(stdout);
		printf( "%d\n", dummy);
		POSITIONS = 10000000;
		REPEATS = 10;
		
		for( long long i = 0; i < POSITIONS; i++) position[ i ] = rand() % countOnes;
		dummy = 0x12345678;
		stoppie = new sdsl::stop_watch();
		stoppie->start();
		for( int k = REPEATS; k-- != 0; )
			for( int i = 0; i < POSITIONS; i += 10 ) {
				dummy ^= ibv.select( position[ i ] );
				dummy ^= ibv.select( position[ i + 1 ] );
				dummy ^= ibv.select( position[ i + 2 ] );
				dummy ^= ibv.select( position[ i + 3 ] );
				dummy ^= ibv.select( position[ i + 4 ] );
				dummy ^= ibv.select( position[ i + 5 ] );
				dummy ^= ibv.select( position[ i + 6 ] );
				dummy ^= ibv.select( position[ i + 7 ] );
				dummy ^= ibv.select( position[ i + 8 ] );
				dummy ^= ibv.select( position[ i + 9 ] );
			}
		stoppie->stop();
		end = (long)std::clock();
		printf("\n%d\t%d\n", start, end);
		s = stoppie->get_user_time() / 1000;
		printf( "%fs, %f selects/s, hsg%fns/select, s3j%f \n", s, (double)(REPEATS * POSITIONS) / s, ((double)(1E9 * s)) / (double)(REPEATS * POSITIONS), (double)ibv.size() / boolvect.size() );fflush(stdout);
		printf( "%d\n", dummy);

	}
#endif /*__MEASURE_RANK__*/
	printf("\n");

/*	#ifdef __MEASURE_SELECT__
	{
		printf("============Testing=======Select============Now======================================== \n");
		#if _FIND_BLOCK_BORDER_RECURSION == _ON
		selectBitVector ibv(bv, _recdepth);
		#else //not RECURSION
		selectBitVector ibv(bv);
		#endif //RECURSION
		printf("\n now testing outside access to selectBitVector, here we go:\n");fflush(stdout);
		unsigned int POSITIONS = 10000000;
		unsigned int REPEATS = 100;
		unsigned int * const position = (unsigned int *)calloc( POSITIONS, sizeof *position );


		for( long long i = POSITIONS; i-- != POSITIONS; ) position[ i ] = rand() % boolvect.size();
		unsigned int dummy = 0x12345678;
		sdsl::stop_watch* stoppie = new sdsl::stop_watch();
		stoppie->start();
		for( int k = REPEATS; k-- != 0; )
			for( int i = 0; i < POSITIONS; i += 10 ) {
				dummy ^= ibv.rank( position[ i ] );
				dummy ^= ibv.rank( position[ i + 1 ] );
				dummy ^= ibv.rank( position[ i + 2 ] );
				dummy ^= ibv.rank( position[ i + 3 ] );
				dummy ^= ibv.rank( position[ i + 4 ] );
				dummy ^= ibv.rank( position[ i + 5 ] );
				dummy ^= ibv.rank( position[ i + 6 ] );
				dummy ^= ibv.rank( position[ i + 7 ] );
				dummy ^= ibv.rank( position[ i + 8 ] );
				dummy ^= ibv.rank( position[ i + 9 ] );
			}
		stoppie->stop();
		end = (long)std::clock();
		printf("\n%d\t%d\n", start, end);
		double s = stoppie->get_user_time() / 1000;
		printf( "%fs, %f ranks/s, hrg%fns/rank, r3j%f \n", s, (double)(REPEATS * POSITIONS) / s, ((double)(1E9 * s)) / (double)(REPEATS * POSITIONS), (double)ibv.size() / boolvect.size() );fflush(stdout);
		printf( "%d\n", dummy);
		POSITIONS = 10000000;
		REPEATS = 100;
		
		for( long long i = POSITIONS; i-- != POSITIONS; ) position[ i ] = rand() % (boolvect.size()/2);
		dummy = 0x12345678;
		stoppie = new sdsl::stop_watch();
		stoppie->start();
		for( int k = REPEATS; k-- != 0; )
			for( int i = 0; i < POSITIONS; i += 10 ) {
				dummy ^= ibv.select( position[ i ] );
				dummy ^= ibv.select( position[ i + 1 ] );
				dummy ^= ibv.select( position[ i + 2 ] );
				dummy ^= ibv.select( position[ i + 3 ] );
				dummy ^= ibv.select( position[ i + 4 ] );
				dummy ^= ibv.select( position[ i + 5 ] );
				dummy ^= ibv.select( position[ i + 6 ] );
				dummy ^= ibv.select( position[ i + 7 ] );
				dummy ^= ibv.select( position[ i + 8 ] );
				dummy ^= ibv.select( position[ i + 9 ] );
			}
		stoppie->stop();
		end = (long)std::clock();
		printf("\n%d\t%d\n", start, end);
		s = stoppie->get_user_time() / 1000;
		printf( "%fs, %f selects/s, hsg%fns/select, s3j%f \n", s, (double)(REPEATS * POSITIONS) / s, ((double)(1E9 * s)) / (double)(REPEATS * POSITIONS), (double)ibv.size() / boolvect.size() );fflush(stdout);
		printf( "%d\n", dummy);
	}
#endif /*__MEASURE_SELECT__*/ 
/*	printf("\n————————————Testing outside select—————————————");
	printf("\n\tlvlASt\tlvlBSt\tlvlBEnd\t\tcode\tblockLength");

	for (unsigned long k = 0; k < 200; k++) {
		printf("\n");
		printf("\tselect %d = %d", k, selector.select(k));
	}//*/

//	printf("sizeof long = %d", sizeof(unsigned long));
/*	indexedBitVector ranker(bv,3);
	for (unsigned int i = 0; i < 300; i++) {
		int bit = bv[i];
		printf("\n%d: rank(%d) = %d",bit, i, ranker.rank(i));
	}
*/	printf("\n");
	return 0;
}

