extern bool debug;

#include "escapeRanker.hpp"

#include <sdsl/bitmagic.hpp>
#include <sdsl/int_vector.hpp>
//#include <sdsl/rank_support_jmc.hpp>

#include <stdio.h>
#include <map>
#include <algorithm>

using namespace std;
using namespace sdsl;

const unsigned long firstKmasks[65] = {	0x00000000, 	0x0000000000000001, 0x0000000000000003, 0x0000000000000007, 0x000000000000000F, 0x000000000000001F, 0x000000000000003F, 0x000000000000007F, 0x00000000000000FF,
							0x00000000000001FF, 0x00000000000003FF, 0x00000000000007FF, 0x0000000000000FFF, 0x0000000000001FFF, 0x0000000000003FFF, 0x0000000000007FFF, 0x000000000000FFFF,
							0x000000000001FFFF, 0x000000000003FFFF, 0x000000000007FFFF, 0x00000000000FFFFF, 0x00000000001FFFFF, 0x00000000003FFFFF, 0x00000000007FFFFF, 0x0000000000FFFFFF,
							0x0000000001FFFFFF, 0x0000000003FFFFFF, 0x0000000007FFFFFF, 0x000000000FFFFFFF, 0x000000001FFFFFFF, 0x000000003FFFFFFF, 0x000000007FFFFFFF, 0x00000000FFFFFFFF,
							0x00000001FFFFFFFF, 0x00000003FFFFFFFF, 0x00000007FFFFFFFF, 0x0000000FFFFFFFFF, 0x0000001FFFFFFFFF, 0x0000003FFFFFFFFF, 0x0000007FFFFFFFFF, 0x000000FFFFFFFFFF,
							0x000001FFFFFFFFFF, 0x000003FFFFFFFFFF, 0x000007FFFFFFFFFF, 0x00000FFFFFFFFFFF, 0x00001FFFFFFFFFFF, 0x00003FFFFFFFFFFF, 0x00007FFFFFFFFFFF, 0x0000FFFFFFFFFFFF,
							0x0001FFFFFFFFFFFF, 0x0003FFFFFFFFFFFF, 0x0007FFFFFFFFFFFF, 0x000FFFFFFFFFFFFF, 0x001FFFFFFFFFFFFF, 0x003FFFFFFFFFFFFF, 0x007FFFFFFFFFFFFF, 0x00FFFFFFFFFFFFFF,
							0x01FFFFFFFFFFFFFF, 0x03FFFFFFFFFFFFFF, 0x07FFFFFFFFFFFFFF, 0x0FFFFFFFFFFFFFFF, 0x1FFFFFFFFFFFFFFF, 0x3FFFFFFFFFFFFFFF, 0x7FFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF};

struct comparator {
	bool operator() (std::pair<unsigned long, unsigned long> i,std::pair<unsigned long, unsigned long> j) {
		return (i.second>j.second);
	}
};

unsigned long escapeRanker::decode(unsigned long i, unsigned long ones) const{
	if (frequentBlocks[i] == 0) return frequentDecoder[ones] [ frequentCodes[i-frequentRank.rank(i)] ];
	else i = frequentRank.rank(i);
	if (rareBlocks[i] == 0) return rareDecoder[ones] [ rareCodes[i - rareRank.rank(i)] ];
	else i = rareRank.rank(i);
	return ultraRare[i];
}

unsigned long escapeRanker::rankA(unsigned long i) const {				//rank information for large block i
	return levelA[i];
}
unsigned long escapeRanker::rankB(unsigned long i) const {				//rank information for small block i
	#ifdef __DEBUG__
	if (debug) {
		printf("rankA(%ld) = %ld, levelB[%ld] = %ld, levelBSize = %ld\n",i/(LEVELA/LEVELB), rankA(i/(LEVELA/LEVELB)), i, levelB[i], levelB.size()); fflush(stdout);
	}
	#endif //debug	low = low * (LEVELA/LEVELB);

	return rankA(i / (LEVELA/LEVELB)) + levelB[i];
}

escapeRanker::escapeRanker(sdsl::int_vector<1> bv){
	build(bv);
}
void escapeRanker::build(sdsl::int_vector<1> bv){
	//INIT
	std::vector<std::map<unsigned long,unsigned long> > patternCount;
	patternCount.resize(LEVELB + 1);
	inputSize = bv.size();
	rareDecoder.resize(LEVELB + 1);
	frequentDecoder.resize(LEVELB + 1);
	unsigned long levelASize = (inputSize - 1) / LEVELA + 1;	//this is the number of blocks, the data structures will have one more element, for watcher reasons
	unsigned long levelBSize = (inputSize - 1) / LEVELB + 1;
	levelA = new unsigned long[levelASize + 1];
	levelB.resize(levelBSize + 1);
	unsigned long* levelC;				//verbatim encoded small blocks
	levelC = new unsigned long[levelBSize];

	for (unsigned long i = 0; i < levelASize+1; i++){
		levelA[i] = 0;
	}
	for (unsigned long i = 0; i < levelB.size(); i++){
		levelB[i] = 0;
	}
	//making blocks
	unsigned long currentLevelA = 0;
	unsigned long r = 0;
	unsigned long currentBlock = bv.get_int(0, LEVELB);
	for (unsigned long i = 0; i < levelBSize; i++) {

//if (i == (8744511/LEVELB + 2)) debug = false;
//if (i == (8744511/LEVELB)) debug = true;
		currentBlock = bv.get_int(i*LEVELB, LEVELB);
		patternCount[bit_magic::b1Cnt(currentBlock)][currentBlock]++;	//was: encoders[bit_magic::b1Cnt(currentBlock)].include(currentBlock);
		if (i*LEVELB%LEVELA == 0){
			levelA[(i*LEVELB)/LEVELA] = r;
			currentLevelA = r;
		}

		levelB[i] = r - currentLevelA;
		levelC[i] = currentBlock;
		r += bit_magic::b1Cnt(currentBlock);
	}
	levelA[levelASize] = r;						//calculate #ones in last block by levelA[i+1] - levelA[i] accesses this element
	if(levelBSize%(LEVELA/LEVELB) == 0) levelB[levelBSize] = 0;	//next small block would be in next large block
	else levelB[levelBSize] = r-currentLevelA;			//calculate #ones in last block by levelB[i+1] - levelB[i] accesses this element

	//order by frequency
	std::vector<sdsl::int_vector<LEVELB> >::iterator frequentIter = frequentDecoder.begin();
	std::vector<sdsl::int_vector<LEVELB> >::iterator rareIter = rareDecoder.begin();
	unsigned int frequentCount = 0, rareCount = 0, ultraRareCount = 0; //how many blocks of which class?

//?	unsigned long compressedSize = 0;
	for (std::vector<std::map<unsigned long,unsigned long> >::iterator iter = patternCount.begin(); iter != patternCount.end(); iter++) {
		//convert hashmap to vector for sorting
//?		unsigned long maxLength = 0;
		std::vector<std::pair<unsigned long, unsigned long> > patternCountVector;
		patternCountVector.resize(iter->size());
		std::vector<std::pair<unsigned long, unsigned long> >::iterator vectIt = patternCountVector.begin();
		std::map<unsigned long, unsigned long>::iterator mapIt;
		for(mapIt = iter->begin(); mapIt != iter->end(); mapIt++) {
//?			unsigned long length = sdsl::bit_magic::l1BP(mapIt->first) + 1;
//?			maxLength = std::max(maxLength, length);
			*vectIt = *mapIt;
			vectIt++;
		}
		//sort
		std::sort(patternCountVector.begin(), patternCountVector.end(), comparator());
//?		decode->init(patternCountVector.size() + 1, maxLength);		//+ 1 is for the zero length encoding we do not want to use

		//convert vector back to hash
		frequentIter->resize(min((size_t)(1 << frequentSize), iter->size()                      ));
		if (iter->size() > 1<<frequentSize) rareIter->resize(min((size_t)(1<<rareSize) , iter->size() - (1<<frequentSize)));
		unsigned long frequentEnd = 0;
		unsigned long rareEnd = 0;
		
		for(unsigned long i = 0; i < patternCountVector.size(); i++){
			(*iter)[patternCountVector[i].first] = i;
			if (i < (1<<frequentSize)){
				(*frequentIter)[frequentEnd++] = patternCountVector[i].first;
				frequentCount += patternCountVector[i].second;
			}
			else if (i < (1<<rareSize)) {
				(*rareIter)[rareEnd++] = patternCountVector[i].first;
				rareCount += patternCountVector[i].second;
			}
			else ultraRareCount++;
		}
		frequentIter++;
		rareIter++;
	}


	//fill levelCcompressed
	frequentCodes.resize(frequentCount);
	rareCodes.resize(rareCount);

	frequentBlocks.resize(levelBSize);
	rareBlocks.resize(levelBSize - frequentCount);
	ultraRare.resize(levelBSize - frequentCount - rareCount);

	unsigned long frequentCodesEnd	= 0;
	unsigned long frequentBlocksEnd	= 0;
	unsigned long rareCodesEnd	= 0;
	unsigned long rareBlocksEnd	= 0;
	unsigned long ultraRareEnd	= 0;
	

	for(unsigned long i = 0; i < levelBSize; i++) {
		unsigned long ones = bit_magic::b1Cnt(levelC[i]);
		unsigned long blockCode = patternCount[ones][levelC[i]];
		if (blockCode < (1<<frequentSize)) {
			frequentCodes[frequentCodesEnd++] = blockCode;
			frequentBlocks[frequentBlocksEnd++] = 0;
		}
		else if (blockCode < (1<<rareSize)) {
			frequentBlocks[frequentBlocksEnd++] = 1;
			rareCodes[rareCodesEnd++] = blockCode - (1<<frequentSize);
			rareBlocks[rareBlocksEnd++] = 0;
		}
		else {		//ultra rare blocks
			frequentBlocks[frequentBlocksEnd++] = 1;
			rareBlocks[rareBlocksEnd++] = 1;
			ultraRare[ultraRareEnd++] = levelC[i];
		}

	}
	frequentRank = nearest_neighbour_dictionary<NND_SAMPLES>(frequentBlocks);
	rareRank = nearest_neighbour_dictionary<NND_SAMPLES>(rareBlocks);
	
	return;
}


unsigned long escapeRanker::rank(unsigned long i) const {
	unsigned long ones, block, beforeBlock, inBlock;
	beforeBlock = rankB(i/LEVELB);
	ones = rankB(i/LEVELB + 1) - beforeBlock;
	block = decode(i/LEVELB, ones);
	inBlock = bit_magic::b1Cnt(block & firstKmasks[(i%LEVELB)]);
	return beforeBlock + inBlock;
}

unsigned long escapeRanker::select(unsigned long i) const {
	unsigned int low = 0, high = (inputSize - 1) / LEVELA + 1, mid = 0;

#ifdef selSampling
//	low = selSamples[i/selSampleDensity];
//	high = selSamples[i/selSampleDensity+1] + 1;
	
#endif

	//binary search on level A
	do {
		mid = (high + low) / 2;
#ifdef __DEBUG__
if (debug) {
	printf("low = %ld, high = %ld mid = %ld\n", low, high, mid); fflush(stdout);
}
#endif //debug
		if (rankA(mid) < i) low = mid;
		else high = mid;

	} while (low + 1 < high);
	#ifdef __DEBUG__
	if (debug) {
		printf("low = %ld, high = %ld\n",low, high); fflush(stdout);
	}
	#endif //debug	//binary search on level B
	i = i - rankA(low);
#ifdef __DEBUG__
if (debug) {
	printf("i = %ld\n", i); fflush(stdout);
}
#endif //debug
	low = low * (LEVELA/LEVELB);
	high = low + (LEVELA/LEVELB);
	if (((inputSize - 1) / LEVELB + 1) < high)
		high = (inputSize - 1) / LEVELB + 1;
	do {
		mid = (high + low) / 2;
#ifdef __DEBUG__
if (debug) {
	printf("low = %ld, high = %ld mid = %ld\n", low, high, mid); fflush(stdout);
}
#endif //debug
		if (levelB[mid] < i) low = mid;
		else high = mid;
	} while (low + 1 < high);
#ifdef __DEBUG__
if (debug) {
	printf("low = %ld, high = %ld\n", low, high); fflush(stdout);
}
#endif //debug	low = low * (LEVELA/LEVELB);

	//decode block
	unsigned long ones, block, beforeBlock, inBlock;
	beforeBlock = rankB(low);
	ones = rankB(low + 1) - beforeBlock;
#ifdef __DEBUG__
if (debug) {
	printf("low = %ld, ones = %ld, beforeBlock = %ld\n", low, ones, beforeBlock); fflush(stdout);
}
#endif //debug
	block = decode(low, ones);
	//select in block
	i = i - levelB[low];
	return low*LEVELB + ((i==0)?0:(sdsl::bit_magic::i1BP(block, i)));

}
unsigned long escapeRanker::getBits(unsigned long i, unsigned long length){		//extract $length bits starting at position $i from the original bitarray
	//TODO
}
unsigned long escapeRanker::size(){
	unsigned int ret = 0;	//measured in bits
	ret += ((inputSize - 1) / LEVELA + 1) * sizeof(unsigned long) * 8;	//levelA
	printf("\nSizes:\nlevelA: %d\n", ((inputSize - 1) / LEVELA + 1) * sizeof(unsigned long) * 8);
	ret += levelB.bit_size();						//levelB
	printf("levelB: %d\n", levelB.size() * logLvlA);
	ret += frequentCodes.bit_size();
	ret += frequentBlocks.bit_size();
	printf("number of Blocks: %d\n", frequentBlocks.bit_size());
	ret += rareCodes.bit_size();
	ret += rareBlocks.bit_size();
	printf("number of rare/ultrarare Blocks: %d\n", rareBlocks.bit_size());
	ret += ultraRare.bit_size();
	printf("space 4 ultraRare: %d\n", ultraRare.bit_size());
	ret += frequentRank.size();	//this is the size of the bitvector itself, not the actual memory usage of the nnd
	ret += rareRank.size();		//as above

	for (std::vector<sdsl::int_vector<LEVELB> >::iterator iter = frequentDecoder.begin(); iter != frequentDecoder.end(); iter++) {
		ret += iter->bit_size();
	}
	for (std::vector<sdsl::int_vector<LEVELB> >::iterator iter = rareDecoder.begin(); iter != rareDecoder.end(); iter++) {
		ret += iter->bit_size();
	}
	printf("HRGR53J: %f,%f\n",((float)(frequentBlocks.bit_size() - rareBlocks.bit_size())) / ((float)frequentBlocks.bit_size()), (float)ultraRare.size()/frequentBlocks.bit_size());
	return ret;	
}
