#ifndef __ESCAPERANKER_HPP_INCLUDED__
#define __ESCAPERANKER_HPP_INCLUDED__

#ifdef escfreq
	#define frequentSize (escfreq)
#else
	#define frequentSize 8
#endif
#ifdef escrare
	#define rareSize (escrare)
#else
	#define rareSize 15
#endif
#define rest LEVELB
#include "constants.hpp"

#include <vector>
#include <sdsl/int_vector.hpp>
#include <sdsl/nearest_neighbour_dictionary.hpp>

class escapeRanker {
	private:
	unsigned long 					inputSize;			//size in bits
	unsigned long*					levelA;				//rank for large block borders
	sdsl::int_vector<logLvlA>			levelB;				//rank for small block borders

	sdsl::int_vector<frequentSize>			frequentCodes;			//frequent blocks encoded
	sdsl::int_vector<1>				frequentBlocks;			//frequent blocks marked with 0
	sdsl::nearest_neighbour_dictionary<NND_SAMPLES>	frequentRank;			//rank support for frequentBlocks
	std::vector<sdsl::int_vector<LEVELB> >		frequentDecoder;		//decode table for frequent blocks

	sdsl::int_vector<rareSize>			rareCodes;			//rare blocks encoded
	sdsl::int_vector<1>				rareBlocks;			//rare blocks marked with 0
	sdsl::nearest_neighbour_dictionary<NND_SAMPLES>	rareRank;			//rank support for rareBlocks
	std::vector<sdsl::int_vector<LEVELB> >		rareDecoder;			//decode table for rare blocks

	sdsl::int_vector<LEVELB>			ultraRare;				//verbatim encoded ultra rare blocks


	private:
	unsigned long rankA(unsigned long i) const;				//rank information for large block i
	unsigned long rankB(unsigned long i) const;				//rank information for small block i
	unsigned long decode(unsigned long i, unsigned long ones) const;		//decode small block number i given the correct number of ones
	public:
	escapeRanker(sdsl::int_vector<1> bv);
	void build(sdsl::int_vector<1> bv);
	unsigned long rank(unsigned long i) const;
	unsigned long select(unsigned long i) const;
	unsigned long getBits(unsigned long i, unsigned long length);		//extract $length bits starting at position $i from the original bitarray
	unsigned long size();
	//TODO? write destructor
};
#endif
