#include "indexedBitVector.hpp"
#include "Practical_ent_rs/combination.hpp"
#include "canonicalEncoder.hpp"
#include "constants.hpp"

#include <sdsl/bitmagic.hpp>
#include <sdsl/int_vector.hpp>
#include <sdsl/select_support_mcl.hpp>
#include <stdio.h>
#include <map>
#include <algorithm>


using namespace std;
using namespace sdsl;

const unsigned long firstKmasks[65] = {	0x00000000, 	0x0000000000000001, 0x0000000000000003, 0x0000000000000007, 0x000000000000000F, 0x000000000000001F, 0x000000000000003F, 0x000000000000007F, 0x00000000000000FF,
							0x00000000000001FF, 0x00000000000003FF, 0x00000000000007FF, 0x0000000000000FFF, 0x0000000000001FFF, 0x0000000000003FFF, 0x0000000000007FFF, 0x000000000000FFFF,
							0x000000000001FFFF, 0x000000000003FFFF, 0x000000000007FFFF, 0x00000000000FFFFF, 0x00000000001FFFFF, 0x00000000003FFFFF, 0x00000000007FFFFF, 0x0000000000FFFFFF,
							0x0000000001FFFFFF, 0x0000000003FFFFFF, 0x0000000007FFFFFF, 0x000000000FFFFFFF, 0x000000001FFFFFFF, 0x000000003FFFFFFF, 0x000000007FFFFFFF, 0x00000000FFFFFFFF,
							0x00000001FFFFFFFF, 0x00000003FFFFFFFF, 0x00000007FFFFFFFF, 0x0000000FFFFFFFFF, 0x0000001FFFFFFFFF, 0x0000003FFFFFFFFF, 0x0000007FFFFFFFFF, 0x000000FFFFFFFFFF,
							0x000001FFFFFFFFFF, 0x000003FFFFFFFFFF, 0x000007FFFFFFFFFF, 0x00000FFFFFFFFFFF, 0x00001FFFFFFFFFFF, 0x00003FFFFFFFFFFF, 0x00007FFFFFFFFFFF, 0x0000FFFFFFFFFFFF,
							0x0001FFFFFFFFFFFF, 0x0003FFFFFFFFFFFF, 0x0007FFFFFFFFFFFF, 0x000FFFFFFFFFFFFF, 0x001FFFFFFFFFFFFF, 0x003FFFFFFFFFFFFF, 0x007FFFFFFFFFFFFF, 0x00FFFFFFFFFFFFFF,
							0x01FFFFFFFFFFFFFF, 0x03FFFFFFFFFFFFFF, 0x07FFFFFFFFFFFFFF, 0x0FFFFFFFFFFFFFFF, 0x1FFFFFFFFFFFFFFF, 0x3FFFFFFFFFFFFFFF, 0x7FFFFFFFFFFFFFFF, 0xFFFFFFFFFFFFFFFF};

struct comparator {
	bool operator() (std::pair<unsigned long, unsigned long> i,std::pair<unsigned long, unsigned long> j) {
		return (i.second>j.second);
	}
};

int enumCode(unsigned long b, int ones, int length, comb& cb){
	int_vector<1> B(length);
	B.set_int(0,b, length);	
	int x = 0;
	for (int i = 0; i < length; i++){
		if (B[i]) {
			x += cb.choose (length - i - 1, ones);
			ones--;
		}
	}
	return x;
}

void output_bitsize(unsigned long x) {
	unsigned long b, B, KiB, MiB, GiB;
	b = x%8;
	x = x/8;
	B = x%1024;
	x = x/1024;
	KiB = x%1024;
	x = x/1024;
	MiB = x%1024;
	x = x/1024;
	GiB = x;
	printf("%4dGiB %4dMiB %4dKiB %4dB %1db", GiB, MiB, KiB, B, b);
}

indexedBitVector::indexedBitVector(int_vector<1> bv) {
	build(bv,0);
}
indexedBitVector::indexedBitVector(int_vector<1> bv, unsigned int recursionDepth) {
	build(bv, recursionDepth);
}

void indexedBitVector::build(int_vector<1> bv, unsigned int recursionDepth) {
	inputSize = bv.size();
	std::vector<canonicalEncoder> encoders;
	encoders.reserve(LEVELB + 1);
	decoders.reserve(LEVELB + 1);
	for (unsigned long i = 0; i < LEVELB + 1; i++) {
		encoders.push_back(canonicalEncoder());
	}
	unsigned long levelASize = (inputSize - 1) / LEVELA + 1;	//this is the number of blocks, the data structures will have one more element, for watcher reasons
	unsigned long levelBSize = (inputSize - 1) / LEVELB + 1;
	printf("levelA = %d, levelB = %d, levelASize = %d levelBSize = %d", LEVELA, LEVELB, levelASize, levelBSize); fflush(stdout);//debugcode
#ifdef selSampling
	selSamples = new unsigned long[(inputSize - 1) / selSampleDensity + 2];
	selSamples[0] = 0;
#endif
	levelA = new unsigned long[levelASize + 1];
	levelB.resize(levelBSize + 1);
	levelC = new unsigned long[levelBSize];
	#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
	levelAptrs = new unsigned long[levelASize + 1];
	levelBptrs.resize(levelBSize + 1);
	#endif	//2LVLPTRS
	for (unsigned long i = 0; i < levelASize+1; i++){
		levelA[i] = 0;
	}
	for (unsigned long i = 0; i < levelBSize+1; i++){
		levelB[i] = 0;
	}

	unsigned long currentLevelA = 0;
	unsigned long r = 0;
	unsigned long currentBlock = bv.get_int(0, LEVELB);
	for (unsigned long i = 0; i < levelBSize; i++) {
		currentBlock = bv.get_int(i*LEVELB, LEVELB);
		encoders[bit_magic::b1Cnt(currentBlock)].include(currentBlock);

		if (i*LEVELB%LEVELA == 0){
			levelA[(i*LEVELB)/LEVELA] = r;
			currentLevelA = r;
		}

		levelB[i] = r - currentLevelA;
		levelC[i] = currentBlock;
#ifdef selSampling
		if (r / selSampleDensity < (r + bit_magic::b1Cnt(currentBlock)) / selSampleDensity) {
			selSamples[(r + bit_magic::b1Cnt(currentBlock))/selSampleDensity] = (i*LEVELB)/LEVELA;
		}
#endif
		r += bit_magic::b1Cnt(currentBlock);
	}
	levelA[levelASize] = r;						//calculate #ones in last block by levelA[i+1] - levelA[i] accesses this element
	if((levelBSize%LEVELA/LEVELB) == 0) levelB[levelBSize] = 0;	//next small block would be in next large block
	else levelB[levelBSize] = r-currentLevelA;			//calculate #ones in last block by levelB[i+1] - levelB[i] accesses this element

	unsigned long compressedSize = 0;	
	for (std::vector<canonicalEncoder>::iterator iter = encoders.begin(); iter != encoders.end(); iter++) {
		iter -> calcCode();
		compressedSize += iter -> predictSize();
		decoders.push_back(iter -> getDecoder());
	} 
	
	printf("\ncompressed Size is %ld:", compressedSize); output_bitsize(compressedSize); printf(" ratio: %f", ((double) compressedSize)/((double) inputSize));
	r = 0;
	//fill levelCcompressed 
	levelCcompressed.resize(compressedSize);
	#if (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL) || (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND)
	levelCbeginnings.resize(compressedSize + 1);		//an additional final 1 to mark the end of the last block
	for(unsigned long i = 0; i < levelCbeginnings.size();i++)
		levelCbeginnings[i] = 0;
	levelCbeginnings[compressedSize] = 1;
	#endif //sdslselect
	for(unsigned long i = 0; i < levelBSize; i++) {
		unsigned long ones = bit_magic::b1Cnt(levelC[i]);
		unsigned long blockCode = encoders[ones].encode(levelC[i]);
		unsigned long blockLength = encoders[ones].getLength();

		levelCcompressed.set_int(r, blockCode, blockLength);
		#if (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL) || (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND)
		levelCbeginnings[r] = 1;
		#endif //_SDSLSELECT
		if (i% (LEVELA/LEVELB) == 0) {
			#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
			levelAptrs[i*LEVELB/LEVELA] = r;
			#endif	//2LVLPTRS
			currentLevelA = r;
		}
		#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
		levelBptrs[i] = r - currentLevelA;
		#endif	//2LVLPTRS
		r+= blockLength;
		#ifdef __DEBUG__
//		if (i > 249999350/LEVELB - 2) debug = true;
		if (debug) {
			printf("%ld\t%lx\t%ld\t%ld\t%ld\n",ones,blockCode,blockLength,r, i*LEVELB);
		}
		debug = false;
		#endif //debug
	}

	#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
	levelAptrs[levelASize] = r;
	if(levelBSize%levelASize == 0) levelBptrs[levelBSize] = 0;		//next small block would be in next large block
	else levelBptrs[levelBSize] = r - currentLevelA;
	#endif	//2LVLPTRS
	#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL
	levelCbeginningsSelector = new select_support_mcl<1,1>(&levelCbeginnings);
	#endif	//SDSLSELECT_MCL
	#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND
	levelCbeginningsSelector = new nearest_neighbour_dictionary<16>(levelCbeginnings);
	#endif //SDSLSELECT_NND
	#if _FIND_BLOCK_BORDER_RECURSION == _ON
	//recursion related stuff
	levelCbeginningsRecursion = new selectBitVector(levelCbeginnings, recursionDepth - 1);
	this->recursionDepth = recursionDepth;
	#endif	//RECURSION
	return;
}

unsigned long indexedBitVector::decode(unsigned long i, unsigned long ones) const{
	unsigned long start, end, length, code;

	#if _FIND_BLOCK_BORDER_RECURSION == _ON
	if (recursionDepth == 0) {
	#endif	//RECURSION
		#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
		start = levelAptrs[i/(LEVELA/LEVELB)] + levelBptrs[i];
		end = levelAptrs[(i+1)/(LEVELA/LEVELB)] + levelBptrs[i+1];
		#endif	//2LVLPTRS
		#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL
		start = levelCbeginningsSelector->select(i+1);
		end = levelCbeginningsSelector->select(i+2);
		#endif //_SDSLSELECT_MCL
		#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND
		start = levelCbeginningsSelector->select(i+1);
		end = levelCbeginningsSelector->select(i+2);
		#endif //_SDSLSELECT_NND
	#if _FIND_BLOCK_BORDER_RECURSION == _ON
	}
	else {
		start = levelCbeginningsRecursion->select(i);
		end = levelCbeginningsRecursion->select(i + 1);
	}
	#endif	//RECURSION
	length = end-start;
	code = levelCcompressed.get_int(start, length);
	#ifdef __DEBUG__
	if (debug) {
		printf("ibv.decode(%ld,%ld): start = %ld, end = %ld, length = %ld code(%%d) = %ld \n", i, ones, start, end, length, code);
		printf("i%%LEVELB = %ld, LEVELB = %d \n", i%LEVELB, LEVELB);
//		printf("levelCBeginnings[%ld] = %d\nlevelCBeginnings[%ld] = %d\nlevelCBeginnings[%ld] = %d\nlevelCBeginnings[%ld] = %d\n",start - 1, levelCbeginnings[start - 1], start, levelCbeginnings[start], start + 1, levelCbeginnings[start+1], start+1, levelCbeginnings[start+1]);
#if (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL) || (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND)
		for (unsigned int i = 0; i < 4000; i++) {
			printf("%d", levelCbeginnings[i]);
			if (i%10==0) printf(" ");
			if (i%100==0) printf("\n");
			if (i%1000==0) printf("\n");
		}
#endif //beginnings
	}
	#endif //debug
	return decoders[ones]->decode(code, length);
}
unsigned long indexedBitVector::rankA(unsigned long i) const{
	return levelA[i];
}
unsigned long indexedBitVector::rankB(unsigned long i) const{
	return rankA(i / (LEVELA/LEVELB)) + levelB[i];
}
unsigned long indexedBitVector::rank(unsigned long i) const{
	unsigned long ones, block, beforeBlock, inBlock;
	beforeBlock = rankB(i/LEVELB);
	ones = rankB(i/LEVELB + 1) - beforeBlock;
	block = decode(i/LEVELB, ones);
	inBlock = bit_magic::b1Cnt(block & firstKmasks[(i%LEVELB)]);
	#ifdef __DEBUG__
	if (debug) {
		printf("ibv.rank(i): beforeBlock = %ld, ones = %ld, block = %lx, inBlock = %ld, return = %ld\n", beforeBlock, ones, block, inBlock, beforeBlock + inBlock);
		printf("i%%LEVELB = %ld, LEVELB = %d \n", i%LEVELB, LEVELB);
	}
	#endif //debug
	return beforeBlock + inBlock;
}
unsigned long indexedBitVector::select(unsigned long i) const{
	unsigned int low = 0, high = (inputSize - 1) / LEVELA + 1, mid = 0;
	#ifdef __DEBUG__
	if (debug) {
		printf("low = %ld, high = %ld, i = %ld\n",low, high, i);
	}
	#endif //debug

#ifdef selSampling
	low = selSamples[i/selSampleDensity];
	high = selSamples[i/selSampleDensity+1] + 1;
	
#endif
	//binary search on level A
	do {
		mid = (high + low) / 2;
		if (rankA(mid) < i) low = mid;
		else high = mid;
		#ifdef __DEBUG__
		if (debug) {
			printf("low = %ld, high = %ld, rankA(%ld)= %ld\n",low, high, mid, rankA(mid));
		}
		#endif //debug
	} while (low + 1 < high);
	//binary search on level B
	i = i - rankA(low);
	low = low * (LEVELA/LEVELB);
	high = low + (LEVELA/LEVELB);
	if (((inputSize - 1) / LEVELB + 1) < high)
		high = (inputSize - 1) / LEVELB + 1;
	#ifdef __DEBUG__
	if (debug) {
		printf("low = %ld, high = %ld, i = %ld\n",low, high, i);
	}
	#endif //debug
	do {
		mid = (high + low) / 2;
		if (levelB[mid] < i) low = mid;
		else high = mid;
		#ifdef __DEBUG__
		if (debug) {
			printf("low = %ld, high = %ld, levelB[%ld] = %ld\n",low, high, mid, levelB[mid]);
		}
		#endif //debug

	} while (low + 1 < high);
	//decode block
	unsigned long ones, block, beforeBlock, inBlock;
	beforeBlock = rankB(low);
	ones = rankB(low + 1) - beforeBlock;
	block = decode(low, ones);
	//select in block
	i = i - levelB[low];
	#ifdef __DEBUG__
	if (debug) {
		printf("ones= %ld, block = %lx, beforeBlock = %ld, inBlock = %ld, i = %ld\n",ones, block, beforeBlock, inBlock, i);
	}
	#endif //debug
	return low*LEVELB + ((i==0)?0:(sdsl::bit_magic::i1BP(block, i)));
/*	{
		unsigned long v = block;          // Input value to find position with rank r.
		unsigned int r = i;      // Input: bit's desired rank [1-64].
		unsigned int s;      // Output: Resulting position of bit with rank r [1-64]
		unsigned long a, b, c, d; // Intermediate temporaries for bit count.
		unsigned int t;      // Bit count temporary.

		// Do a normal parallel bit count for a 64-bit integer,                     
		// but store all intermediate steps.                                        
		// a = (v & 0x5555...) + ((v >> 1) & 0x5555...);
		a =  v - ((v >> 1) & ~0UL/3);
		// b = (a & 0x3333...) + ((a >> 2) & 0x3333...);
		b = (a & ~0UL/5) + ((a >> 2) & ~0UL/5);
		// c = (b & 0x0f0f...) + ((b >> 4) & 0x0f0f...);
		c = (b + (b >> 4)) & ~0UL/0x11;
		// d = (c & 0x00ff...) + ((c >> 8) & 0x00ff...);
		d = (c + (c >> 8)) & ~0UL/0x101;
		t = (d >> 32) + (d >> 48);
		// Now do branchless select!                                                
		s  = 64;
		s -= ((t - r) & 256) >> 3; r -= (t & ((t - r) >> 8));
		t  = (d >> (s - 16)) & 0xff;
		s -= ((t - r) & 256) >> 4; r -= (t & ((t - r) >> 8));
		t  = (c >> (s - 8)) & 0xf;
		s -= ((t - r) & 256) >> 5; r -= (t & ((t - r) >> 8));
		t  = (b >> (s - 4)) & 0x7;
		s -= ((t - r) & 256) >> 6; r -= (t & ((t - r) >> 8));
		t  = (a >> (s - 2)) & 0x3;
		s -= ((t - r) & 256) >> 7; r -= (t & ((t - r) >> 8));
		t  = (v >> (s - 1)) & 0x1;
		s -= ((t - r) & 256) >> 8;
		s = 65 - s;
		return low*LEVELB + s;
	}/**/
}


unsigned long indexedBitVector::getBits(unsigned long i, unsigned long length) {
	unsigned int blockNr, offset;
	blockNr = i / LEVELB;
	offset = i % LEVELB;
	if (offset + length <= LEVELB) {			//querry within one block
		unsigned long ones, block;
		ones = rankB(blockNr + 1) - rankB(blockNr);
		block = decode(blockNr, ones);
		return (block >> offset) & firstKmasks[length];
	}
	else {								//querry over two blocks
		unsigned long ones1, ones2, block1, block2, length1, length2, ret;
		ones1 = rankB(blockNr + 1) - rankB(blockNr);
		block1 = decode(blockNr, ones1);
		ones2 = rankB(blockNr + 2) - rankB(blockNr + 1);
		block2 = decode(blockNr + 1, ones2);
		length1 = LEVELB - offset;
		length2 = length - length1;
		ret =  (block2 & firstKmasks[length2]) << length1;
		return ((block1 >> i) & firstKmasks[length1]) | ret;
	}
}

unsigned long indexedBitVector::size() {
	unsigned long ret = 0;
	ret += ((inputSize - 1) / LEVELA + 1) * sizeof(unsigned long) * 8;	//levelA
	printf("\nSizes:\nlevelA: %d\n", ((inputSize - 1) / LEVELA + 1) * sizeof(unsigned long) * 8);
	ret += levelB.size() * logLvlA;						//levelB
	printf("levelB: %d\n", levelB.size() * logLvlA);
	ret += levelCcompressed.size();						//levelC
	printf("levelC: %d\n", levelCcompressed.size());
	for (vector<decoder*>::iterator iter = decoders.begin(); iter != decoders.end(); iter++) {	//decoders
		ret += (*iter)->size() * 8;
	}
	printf("A+B+C+decoders: %d\n", ret);

	#if _FIND_BLOCK_BORDER_RECURSION == _ON
	printf("rec\n");
	if (recursionDepth == 0) {
	#endif	//RECURSION

		#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
		ret += ((inputSize - 1) / LEVELA + 2) * sizeof(unsigned long) * 8;	//levelAptrs
		printf("levelAptrs: %d\n", ((inputSize - 1) / LEVELA + 2) * sizeof(unsigned long) * 8);
		ret += levelBptrs.size() * logLvlA;					//levelBptrs
		printf("levelBptrs: %d\n", levelBptrs.size() * logLvlA);
		#endif	//2LVLPTRS
		#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL
		ret += sizeof(*levelCbeginningsSelector) * 8;
		#endif //SDSLSELECT_MCL
		#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND
		ret += sizeof(*levelCbeginningsSelector) * 8;
		#endif //SDSLSELECT_MCL

	#if _FIND_BLOCK_BORDER_RECURSION == _ON
	}
	else {
		int rec = levelCbeginningsRecursion->size();
		ret += rec;
		printf("rec: %d", rec);
	}
	#endif	//RECURSION
	#ifdef selSampling
	ret +=  sizeof(unsigned long) * 8 * ((inputSize - 1) / selSampleDensity + 2);
	#endif
	return ret;
}

//TODO write destructor to prevent memory leak (destructors)
