#ifndef __INDEXEDBITVECTOR_HPP_INCLUDED__
#define __INDEXEDBITVECTOR_HPP_INCLUDED__
#include "constants.hpp"
#include <vector>
#include <sdsl/int_vector.hpp>
#include <sdsl/nearest_neighbour_dictionary.hpp>
#include "decoder.hpp"
#include "selectBitVector.hpp"
class indexedBitVector {
	private:
	#ifdef selSampling
	unsigned long*				selSamples;
	#endif
	unsigned long 				inputSize;			//size in bits
	unsigned long*				levelA;				//rank for large block borders
	sdsl::int_vector<logLvlA>		levelB;				//rank for small block borders
	unsigned long*				levelC;				//verbatim encoded small blocks
	sdsl::int_vector<1>			levelCcompressed;		//compressed encoded small blocks
	#if ((_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL) || (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND) || (_FIND_BLOCK_BORDER_RECURSION == _ON))
	sdsl::int_vector<1>			levelCbeginnings;		//beginnings of compressed encoded small blocks indicated by a 1
	#endif //SELECT or RECURSION

	#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
	unsigned long*				levelAptrs;			//beginnings of large blocks in compressed small block encoding
	sdsl::int_vector<logLvlA>		levelBptrs;			//beginnings of small blocks in compressed small block encoding relative to large block beginnings
	#endif	//2LVLPTRS
	#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL
	sdsl::select_support_mcl<1,1>*		levelCbeginningsSelector;
	#endif //_SDSLSELECT_MCL
	#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND
	sdsl::nearest_neighbour_dictionary<NND_SAMPLES>* levelCbeginningsSelector;
	#endif //_SDSLSELECT_NND

	#if _FIND_BLOCK_BORDER_RECURSION == _ON
	unsigned int				recursionDepth;			//depth of further recursion, 0 is at abbortlevel
	selectBitVector*			levelCbeginningsRecursion;	//the recursive way to obtain code beginnings
	#endif	//RECURSION

	std::vector<decoder*>			decoders;
	
	private:
	unsigned long rankA(unsigned long i) const;				//rank information for large block i
	unsigned long rankB(unsigned long i) const;				//rank information for small block i
	unsigned long decode(unsigned long i, unsigned long ones) const;		//decode small block number i given the correct number of ones
	public:
	indexedBitVector(sdsl::int_vector<1> bv);
	indexedBitVector(sdsl::int_vector<1> bv, unsigned int recursionDepth);
	void build(sdsl::int_vector<1> bv);
	void build(sdsl::int_vector<1> bv, unsigned int recursionDepth);
	unsigned long rank(unsigned long i) const;
	unsigned long select(unsigned long i) const;
	unsigned long getBits(unsigned long i, unsigned long length);		//extract $length bits starting at position $i from the original bitarray
	unsigned long size();
	//TODO write destructor to clear up decoders
};
#endif
