#ifndef __SELECTBITVECTOR_HPP_INCLUDED__
#define __SELECTBITVECTOR_HPP_INCLUDED__
#include "constants.hpp"

#include <vector>
#include <sdsl/int_vector.hpp>
#include <sdsl/nearest_neighbour_dictionary.hpp>
#include <sdsl/rank_support.hpp>

#include "decoder.hpp"


class selectBitVector {
	private:
	#ifdef rnkSampling //NYI
	unsigned long*				rnkSamples;
	#endif
	unsigned long 					inputSize;			//size in bits
	unsigned long					totalOnes;			//total number of ones
	unsigned long*					levelA;				//select for large block borders
	sdsl::int_vector<1>				levelASparseIndicator;		//sparse blocks indicated by a 1
	sdsl::nearest_neighbour_dictionary<NND_SAMPLES>	levelASparseRanker;
	sdsl::int_vector<logSelLvlA>			levelB;				//select for small block borders
	sdsl::int_vector<1>				levelBSparseIndicator;		//sparse blocks indicated by a 1
	sdsl::nearest_neighbour_dictionary<NND_SAMPLES>	levelBSparseRanker;
	unsigned long*					levelC;				//verbatim encoded small blocks
	std::vector<std::vector<unsigned long> >	levelASparseTable;		//table for sparse large blocks
	std::vector<sdsl::int_vector<logSelLvlA> >	levelBSparseTable;		//table for sparse small blocks

	sdsl::int_vector<1>			levelCcompressed;		//compressed encoded small blocks
	#if ((_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL) || (_FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND) || (_FIND_BLOCK_BORDER_RECURSION == _ON))
	sdsl::int_vector<1>			levelCbeginnings;		//beginnings of compressed encoded small blocks indicated by a 1
	#endif //SELECT or RECURSION	selectBitVector*			levelCbeginningsRecursion;
	unsigned int				recursionDepth;
	#if _FIND_BLOCK_BORDER_MODE == _2LVLPTRS
	unsigned long*				levelAptrs;			//beginnings of large blocks in compressed small block encoding
	sdsl::int_vector<logSelLvlA>		levelBptrs;			//beginnings of small blocks in compressed small block encoding relative to large block beginnings
	#endif //_2LVLPTRS
	#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_MCL
	sdsl::select_support_mcl<1,1>*		levelCbeginningsSelector;
	#endif //_SDSLSELECT_MCL
	#if _FIND_BLOCK_BORDER_MODE == _SDSLSELECT_NND
	sdsl::nearest_neighbour_dictionary<NND_SAMPLES>* levelCbeginningsSelector;
	#endif //_SDSLSELECT_NND
	std::vector<decoder*>			decoders;

	private:
	unsigned long decode(unsigned long i,unsigned long blockLength) const;	//extract code of small block number i given its length

	public:
	selectBitVector(){};
	selectBitVector(sdsl::int_vector<1> bv);
	selectBitVector(sdsl::int_vector<1> bv, unsigned int recursionDepth);
	void build(sdsl::int_vector<1> bv);
	void build(sdsl::int_vector<1> bv, unsigned int recursionDepth);
	unsigned long select(unsigned long i) const;
	unsigned long rank(unsigned long i) const;
	unsigned long size();
};
#endif
