#include "tableDecoder.hpp"
#include "decoder.hpp"
#include "constants.hpp"
#include <iostream>

tableDecoder::tableDecoder(unsigned long size, unsigned long maxLength) {
	init(size, maxLength);
}

tableDecoder::tableDecoder() {
	this->maxLength = 2;
	//everything initiated in init
}

void tableDecoder::init(unsigned long size, unsigned long maxLength) {
	this->maxLength = maxLength;
	decodeTable.set_int_width(maxLength);
	decodeTable.resize(size);
	return;
}

bool tableDecoder::add(unsigned long code, unsigned long length, unsigned long decode){
	unsigned long index = ((1 << length) + code - 1);
	if (index >= decodeTable.size()) return false;
	decodeTable[index] = decode;
//	printf("\t(%2d,%2d,%8x)", code, length, decode);
	return true;
}

unsigned long tableDecoder::decode(unsigned long code, unsigned long length){
	#ifdef __DEBUG__
	if (debug) {
		unsigned long temp = decodeTable[(( 1 << length) + code - 1)];
		printf("tableDecoder.decode(%lx, %ld): return = %llx, maxlength = %ld, decodetable.int_width = %lld\n", code, length, temp, maxLength, decodeTable.get_int_width());
		for (unsigned long i = 0; (i < decodeTable.size()) &&(i < 20); i++) {
			temp = decodeTable[i];
			printf("%d: %llx\n",i,  temp);
		}
	}
	#endif //debug
	return decodeTable[((1 << length) + code - 1)];

}

unsigned long tableDecoder::size() {
	return decodeTable.size() * maxLength / 8;
}
