#ifndef __TABLEDECODER_HPP_INCLUDED__
#define __TABLEDECODER_HPP_INCLUDED__

#include "decoder.hpp"
#include <sdsl/int_vector.hpp>


class tableDecoder : public decoder {
	private:
		unsigned long maxLength;						//maximum length of decoded data in bits
		sdsl::int_vector<0> decodeTable;				//will be set to maxLength in the constructor
		

	public:
		/** constructor with initiatisation
		*
		* @param size: The number of codes the decoder will have to store.
		* @param maxLength: The maximum length of decoded bitpatterns in bits.
		*/
		tableDecoder(unsigned long size, unsigned long maxLength);

		/** default constructor
		* @warning invoke init before use if constructed that way
		*/
		tableDecoder();

		/** initialize data when constructing with default constructor
		*
		* @param size: The number of codes the decoder will have to store.
		* @param maxLength: The maximum length of decoded bitpatterns in bits.
		*/
		void init(unsigned long size, unsigned long maxLength);

		/** add element to decode
		*
		* @param code: the code used to access the bitpattern
		* @param length: the length of the code
		* @param decode: the bitpattern that shall be accessed via the code
		* @return indicates success
		*/
		bool add(unsigned long code, unsigned long length, unsigned long decode);

		/** decode a code
		*
		* @param code: the code to decode
		* @param length: the length of the code
		* @return the decoded bitpattern
		*/ 
		unsigned long decode(unsigned long code, unsigned long length);

		/** get total size of the data struktures in bytes
		*
		* @return, size in bytes
		*/
		unsigned long size();
};
#endif
