#include <iostream>
#include <math.h>
#include "bitVector.hpp"
#include "combination.hpp"
#include "verticalCode.hpp"
#include "appLog.hpp"
#include "common.hpp"

#define logM 5
#define M (1<<logM) //tiny block
#define logP 8
#define P (1<<logP)	//medium block
#define logL 12
#define L (1<<logL) //largeBlock

#define logS 5		//<= 5
#define S (1<<logS) //select block 

#define APPMAPSIZE (L)

#define NON_VC
//#define EN_DIF
#define PACK
#define SELECTNORMAL

static appLog al(APPMAPSIZE);

static const uint widthLg[] = {
  0,5,9,13,16,18,20,22,24,25,26,27,28,29,29,30,30,30,29,29,28,27,26,25,24,22,20,18,16,13,9,5,0};


float entropy(const uint s1, const uint s2){
	float total = (float)(s1+s2);
	return (s1*log(total/s1)+s2*log(total/s2))/log(2.f);
}

uint estimate(const uint sum, const uint total){
	if (sum == 0)  return 0; // total / M;
	else return (uint)entropy(sum,total-sum) + 1; // + total/M;
}

uint appEntropy(const uint s1, const uint s2, const appLog& al){
	uint total = s1+s2;
	return (s1*al.get(s1,total)+s2*al.get(s2,total)) >> FP;
}

uint appEstimate(const uint sum, const uint total, const appLog& al){
	if (sum == 0 || sum == total) return 0;
	else { 
		return appEntropy(sum,total-sum,al);// + total/P;
		//return (uint)entropy(sum,total-sum) + 1;
	}
}

uint lg(uint n){
	uint x = 0;
	while (n){
		x++;
		n>>=1;
	}
	return x;
}

void printBits(uint x){
	for (uint i = 0; i < 32; i++){
		if ((x >> i) & 1) putchar('1');
		else              putchar('0');
		if ((i+1) % 8 == 0) putchar(' ');
	}
	printf("\n");
}

uint enumCode(const bitVector& bv, const comb& cb, uint oneNum, const uint num, const uint i){
	uint x = 0;
	for (uint k = 0; k < num; k++){
		if (i+k < bv.getSize() && bv.getBit(i+k)){
			x += cb.choose(oneNum,num-k-1);
			oneNum--;
		}
	}
	return x;
}			

/*
uint enumDecode(const comb& cb, uint oneNum, uint num, uint x){
	uint bits = 0;
	for (uint k = 0; k < num; k++) {
		if (x >= cb.choose(oneNum,num-k-1)){
			x -= cb.choose(oneNum,num-k-1);
			if (oneNum == 0){
				fprintf(stderr,"enumDecode error\n");
				return 0;
			}
			oneNum--;
			bits |=  (1U << (32 - 1 - k)); //32 should be replaced
		}
	}
	return bits;
}
*/

uint popCount32(const uint x){
	return	popCount[x         & 0xFF]+
			popCount[(x >> 8 ) & 0xFF]+
			popCount[(x >> 16) & 0xFF]+
			popCount[(x >> 24) & 0xFF];
}

class packedBV{
public:
	packedBV():packSize(0){}
	packedBV(uint* s, uint n):packSize(0){
		build(s,n);
	}

	packedBV(ushort* s, uint n):packSize(0){
		build(s,n);
	}

	~packedBV() {}

	void build(uint* s, uint n){
		uint maxS = 0;
		for (uint i = 0; i < n; i++){
			if (s[i] > maxS){
				maxS = s[i];
			}
		}
		packSize = lg(maxS);
		bv.resize(packSize*n);
		for (uint i = 0; i < n; i++){
			bv.setBits(i*packSize,packSize,s[i]);
		}
	}

	void build(ushort* s, uint n){
		uint maxS = 0;
		for (uint i = 0; i < n; i++){
			if (s[i] > maxS){
				maxS = s[i];
			}
		}
		packSize = lg(maxS);
		bv.resize(packSize*n);
		for (uint i = 0; i < n; i++){
			bv.setBits(i*packSize,packSize,s[i]);
		}
	}
	
	uint operator [](const uint x) const{
		return bv.getBits(x*packSize,packSize);
	}


	uint getSize() const{
		return bv.getSize();
	}

	uint getPackSize() const{
		return packSize;
	}

	void write(FILE* outfp){
		fwrite(&packSize,sizeof(uint),1,outfp);
		bv.write(outfp);
	}

	void read(FILE* infp){
		fread(&packSize,sizeof(uint),1,infp);
		bv.read(infp);
	}

public:


private:
	bitVector bv;
	uint packSize;
};

class compBV{
public:
	compBV():cb(M){}
	compBV(bitVector& bv):cb(M) {build(bv);}
	void build(bitVector& bv){
		n = bv.getSize();

		blockL = (n + L - 1)/L; //number of large blocks
		blockP = (n + P - 1)/P; //number of medium blocks
		blockM = (n + M - 1)/M; //number of tiny blocks

		uint* _largePtr   = new uint [blockL];
		uint* _largeRank  = new uint [blockL];
		uint* _vcOffset   = new uint [blockL];
		ushort* _mediumRank = new ushort [blockP];
		uint* _tinyRank   = new uint [blockM];
		uint* _tinyBsum   = new uint [blockM];
		uint* _mediumMmsb = new uint [blockP];
		
		uint* _mediumPtr_dif = new uint [blockM];

		uchar* _tinyVC = new uchar[(n+8-1)/8];

		bitVector _buf((n*2)*8);
		bitVector _largeBuf(L*8);

		uint sl = 0;
		uint pl = 0;
		uint _bufPos = 0;
		uint _bufPosLast = 0;
		realBufSize = 0;
		vcSize = 0;

		verticalCode tinyVC;
		appLog al(APPMAPSIZE);
		
		for (uint il = 0; il < n; il += L){ //Large Block
			_largeRank[il/L] = sl;
			_largePtr [il/L] = _bufPosLast;

			uint sm = 0;
			uint pm = 0;
			uint cumMmsb = 0;
			uint _largeBufPosLast = 0;


			//--debug
			uint preDif = 0;

			for (uint im = 0; im < L && il + im < n; im += P){ //Medium block		
				uint ss = 0;
				uint ps = 0;
				_mediumRank[(il+im)/P] = sm;
				//printf("%d\n",sm);
				
				//uint estimatedPtr_medium = estimate(sm,im);
				uint estimatedPtr_medium  = appEstimate(sm,im,al);
				_mediumPtr_dif[(il+im)/P] = estimatedPtr_medium-pm;

				if (estimatedPtr_medium < _largeBufPosLast){
					
				}
				//printf("%d (est:%d act:%d opt:%d) sm:%d im:%d\n",_largeBufPosLast-estimatedPtr_medium,estimatedPtr_medium,_largeBufPosLast,pm,sm,im);
				
			//	printf("[%d %d %d]\n",(il+im)/P,_mediumPtr_dif[(il+im)/P],estimatedPtr_medium);
				
#ifdef EN_DIF
			//	printf("different %d %d\n",estimatedPtr_medium,pm);
				estimatedPtr_medium = pm;
				
#endif

				//printf("%d:%d \n",(il+im)/P,_mediumPtr_dif[(il+im)/P]);
			
				//	printf("dif:%d t_dif:%d estimate:%d actual:%d\n",estimatedPtr_medium-pm,estimatedPtr_medium-pm-preDif,estimatedPtr_medium,pm);
				//	preDif = estimatedPtr_medium-pm;
				
				//printf("s:%u i:%u ent:%d app:%d\n",sm,im,estimatedPtr_medium,appEstimatedPtr_medium);

				for (uint is = 0; is < P && il + im + is < n; is += M){ //Tiny block
					uint bsum = 0;
					for (uint j = 0; j < M && il + im + is + j < n; j++){
						bsum += bv.getBit(il + im + is + j);
					}
					uint x = enumCode(bv,cb,bsum,M,il+im+is);
					
					uint width = lg(cb.choose(bsum,M)-1);
					
					//uint estimatedPtr  = estimate(ss,is);
					uint estimatedPtr = appEstimate(ss,is,al);

					if (estimatedPtr < ps){
						printf("error! %d %d\n",estimatedPtr_medium,pm);
					}
					
					//printf("s:%u i:%u ent:%d app:%d\n",sm,im,estimatedPtr,appEstimatedPtr);
					
				//	printf("estimatedPtr_medium:%d\n",estimatedPtr_medium);

					
					if (width != 0){
						//_buf.setBits(_largeBufPos+estimatedPtr_medium+estimatedPtr,width,x);
						_largeBuf.setBits(estimatedPtr_medium+estimatedPtr,width,x);
					}


					_largeBufPosLast = estimatedPtr_medium+estimatedPtr+width;
					//printf("[%d %d %d %d %d]\n",_bufPosLast,_largeBufPos,estimatedPtr_medium,estimatedPtr,width);

					//printf("setPos:%d width:%d sm:%d im:%d ss:%d is:%d bsum:%d x:%d\n",estimatedPtr_medium+estimatedPtr,width,sm,im,ss,is,bsum,x);
					ss += bsum;
					ps += width;

					_tinyRank [(il+im+is)/M] = ss;
					_tinyBsum [(il+im+is)/M] = bsum;
					tinyVC.add(bsum);
					//printf("%d %d\n",il+im+is,bsum);
				}
				tinyVC.convert();
				for (uint i = 0; i < tinyVC.getMmsb(); i++){
					unsigned int x = tinyVC.getCode(i);
					_buf.setBits(_bufPosLast+(i*8),8,x);
				}
				_bufPosLast += (tinyVC.getMmsb())*8;
				vcSize += tinyVC.getMmsb();
				{
					_mediumMmsb[(il+im)/P] = tinyVC.getMmsb() + cumMmsb;
					cumMmsb += tinyVC.getMmsb();
				}
				tinyVC.reset();

				sm += ss;
				pm += ps;				
			}
			//printf("offset:%d %d %d\n",_bufPosLast-_largePtr[il/L],cumMmsb*8,_mediumMmsb[(il+L)/P-1]*8);
			//printf("getBuf(8705):%d\n",_largeBuf.getBit(8705));
			_buf.copy(_largeBuf,_bufPosLast,_largeBufPosLast);
			_vcOffset[il/L] = _bufPosLast - _largePtr[il/L];
			_bufPosLast += _largeBufPosLast;
			realBufSize += _largeBufPosLast;

			sl += sm;
			pl += pm;
		}
		sum = sl;
		/*
		for (uint i = 0; i < blockL; i++){
			printf("%d %d\n",_largePtr[i],_largeRank[i]);
		}
		for (uint i = 0; i < blockP; i++){
			printf("%d\n",_mediumRank[i]);
		}
		for (uint i = 0; i < blockM; i++){
			printf("%d\n",_tinyRank[i]);
		}
		*/		

		vcOffset.build(_vcOffset,blockL);
		
#ifdef PACK
		largePtr.build(_largePtr,blockL);
		largeRank.build(_largeRank,blockL);
		mediumRank.build(_mediumRank,blockP);
		delete[] _mediumRank;
		delete[] _largePtr;
		delete[] _largeRank;
#else
		mediumRank = _mediumRank;
		largePtr = _largePtr;
		largeRank = _largeRank;
#endif
		mediumPtr_dif.build(_mediumPtr_dif,blockP);
		mmsbV.build(_mediumMmsb,blockP);
		tinyRank.build(_tinyRank,blockM);
		buf.resize(_bufPosLast);
		buf.copy(_buf,0,_bufPosLast);

		setCount = sl;
		optSize  = pl;
		
		if (_vcOffset) delete[] _vcOffset;
		if (_tinyRank) delete[] _tinyRank;
		if (_tinyBsum) delete[] _tinyBsum;
		if (_mediumMmsb) delete[] _mediumMmsb;
		if (_mediumPtr_dif) delete [] _mediumPtr_dif;
	}

	compBV(FILE* infp):cb(M){
		fread(&n,sizeof(uint),1,infp);
		fread(&setCount,sizeof(uint),1,infp);
		fread(&optSize,sizeof(uint),1,infp);
		fread(&realBufSize,sizeof(uint),1,infp);
		fread(&vcSize,sizeof(uint),1,infp);
		blockL = (n + L - 1)/L;
		blockP = (n + P - 1)/P;
		blockM = (n + M - 1)/M; //number of tiny blocks
		buf.read(infp);	
		vcOffset.read(infp);
#ifdef PACK
		largePtr.read(infp);
		largeRank.read(infp);
		mediumRank.read(infp);
#else
		largePtr = new uint [blockL];
		largeRank = new uint [blockL];
		mediumRank = new ushort [blockP];
		fread(largePtr,sizeof(uint),blockL,infp);
		fread(largeRank,sizeof(uint),blockL,infp);
		fread(mediumRank,sizeof(ushort),blockP,infp);
#endif
		mediumPtr_dif.read(infp);
		mmsbV.read(infp);

		//check
		

		/*
		printf("getTinyRank\n");
		for(uint i = 0; i < n; i+= M){
			uint bsum,rank;
			getTinyRank(i,rank,bsum);
			printf("%d %d %d\n",i,rank,bsum);
		}
		
		printf("mmsbV\n");
		for(uint i = 0; i < blockP; i++){
			printf("%d %d\n",i,mmsbV.get(i));
		}
		*/

		/*
		for (uint i = 0; i < blockP; i++){
			printf("[%d %d]\n",i,mediumPtr_dif.get(i));
		}
		*/

		//tinyRank.read(infp);
		//printf("tinyRank packSize:%d\n",tinyRank.getPackSize());
	}

	void write(FILE* outfp){
		fwrite(&n,sizeof(uint),1,outfp);
		fwrite(&setCount,sizeof(uint),1,outfp);
		fwrite(&optSize,sizeof(uint),1,outfp);
		fwrite(&realBufSize,sizeof(uint),1,outfp);
		fwrite(&vcSize,sizeof(uint),1,outfp);

		buf.write(outfp);
		
		//largePtr.write(outfp);
		//largeRank.write(outfp);
		vcOffset.write(outfp);
#ifdef PACK
		largePtr.write(outfp);
		largeRank.write(outfp);
		mediumRank.write(outfp);
#else
		fwrite(largePtr,sizeof(uint),blockL,outfp);
		fwrite(largeRank,sizeof(uint),blockL,outfp);
		fwrite(mediumRank,sizeof(ushort),blockP,outfp);
#endif
		//mediumRank.write(outfp);
		mediumPtr_dif.write(outfp);
		//tinyRank.write(outfp);
		mmsbV.write(outfp);

		//buf.printBits();
		//printf("\n");
	}

	void printAllocStatus() const{
#ifdef PACK
		uint mediumRankSize = mediumRank.getSize();
		uint largePtrSize   = largePtr.getSize();
		uint largeRankSize  = largeRank.getSize();
#else
		uint mediumRankSize = sizeof(ushort)*blockP*8;
		uint largePtrSize = sizeof(uint)*blockL*8;
		uint largeRankSize = sizeof(uint)*blockL*8;
#endif

		printf("n=%d m=%d p=%f\n",n,setCount,(float)setCount/n);
		printf("entrop:%8.f\n"
               "   opt:%8d\n",entropy(setCount,n-setCount),optSize);
		printf(" total:%8d\n"
               "vector:%8d\n"
               "    LR:%8d\n"
               "    LP:%8d\n"
               "    MR:%8d\n"
               "    VC:%8d\n"
               " MMSBV:%8d\n",
			buf.getSize()+largePtrSize+largeRankSize+mediumRankSize+mmsbV.getSize()
			,realBufSize,largeRankSize,largePtrSize,mediumRankSize,vcSize*8,mmsbV.getSize());
		//	 printf("--- Err:%d *Srank:%d *MptrDif:%d\n",buf.getSize()-realBufSize-vcSize*8,tinyRank.getSize(),mediumPtr_dif.getSize());
	}

	void getTinyRank(uint pos, uint& rank, uint& bsum) const{
		const uint largePos   = pos / L;
		const uint mediumPos  = pos / P;

		const uint lp = largePtr[largePos];

		//retrieve mmsb
		uint mmsbCum = 0;
		if (mediumPos % (L/P) != 0){
			mmsbCum = mmsbV[mediumPos-1];
		}
		uint mmsb    = mmsbV[mediumPos]-mmsbCum;
		uint offset  = (pos % P) / M; //0...7
		//printf("offset:%d mmsb:%d\n",offset,mmsb);
		if (offset == 0){
			rank = 0;
			bsum = 0;
			for (uint i = 0; i < mmsb; i++){
				//printf("buf:%d\n",buf.getBits(lp+(mmsbCum+i)*8,8));
				bsum += popCount[buf.getBits(lp+(mmsbCum+i)*8,8) & mask_l[offset]] << i;
			}
		}
		else {
			rank = 0;
			bsum = 0;
			for (uint i = 0; i < mmsb; i++){
				unsigned int x = buf.getBits(lp+(mmsbCum+i)*8,8);
				//printf("buf:%d\n",x);
				rank += popCount[x & mask_l[offset-1]] << i;
				bsum += popCount[x & mask_l[offset  ]] << i;
			}
			bsum -= rank;
		}
	}

	uint getBit(uint pos) const{
		const uint largePos   = pos / L;
		const uint mediumPos  = pos / P;
		const uint tinyPos    = pos / M;

		const uint lp = largePtr[largePos];
		const uint ls = largeRank[largePos];

		const uint ms = mediumRank[mediumPos];

		uint ss, bsum;
		getTinyRank(pos,ss,bsum);
		
		//const uint  estimatedPtr_medium = estimate(ms,mediumPos * P % L);
		//const uint  estimatedPtr_tiny   = estimate(ss,tinyPos * M % P);
		
#ifdef EN_DIF
		const uint  estimatedPtr_medium = appEstimate(ms,mediumPos * P % L,al) - mediumPtr_dif.get(mediumPos);
#else
		const uint  estimatedPtr_medium = appEstimate(ms,mediumPos * P % L,al);
#endif

		const uint  estimatedPtr_tiny   = appEstimate(ss,tinyPos * M % P,al);

		if (bsum > 32){
			printf("error pos:%d bsum:%d\n",pos,bsum);
			exit(1);
		}
		const uint width = widthLg[bsum];
		const uint skip = vcOffset[largePos];
		uint x;
		if (width == 0) {
			if (bsum == M) x = (uint)(-1);
			else           x = 0;
		}
		else            x  = cb.enumDecode(bsum,M,buf.getBits(lp+skip+estimatedPtr_medium + estimatedPtr_tiny,width));

		//printBits(x);
	
		const uint offset = pos % M;

		//printf("pos:%d setPos:%d width:%d sm:%d im:%d ss:%d is:%d bsum:%d x:%d vcOffset:%d\n",pos,lp+estimatedPtr_medium + estimatedPtr_tiny,width,ms,mediumPos * P % L,ss,tinyPos * M % P,bsum,buf.getBits(lp+skip+estimatedPtr_medium + estimatedPtr_tiny,width),skip);
		//printf("pos:%d lp:%d ls:%d mp:%d ms:%d sp:%d ss:%d bsum:%d width:%d x:%d\n",pos,lp,ls,estimatedPtr_medium,ms,estimatedPtr_tiny,ss,bsum,width,x)
	
		return x >> (M-offset-1) & 1;
	}

	uint rank(uint pos) const{
		const uint largePos   = pos / L;
		const uint mediumPos  = pos / P;
		const uint tinyPos    = pos / M;

		const uint lp = largePtr[largePos];
		const uint ls = largeRank[largePos];
		const uint ms = mediumRank[mediumPos];
		uint ss, bsum;
		getTinyRank(pos,ss,bsum);
		
		//const uint  estimatedPtr_medium = estimate(ms,mediumPos * P % L);
		//const uint  estimatedPtr_tiny   = estimate(ss,tinyPos * M % P);

#ifdef EN_DIF
		const uint  estimatedPtr_medium = appEstimate(ms,mediumPos * P % L,al) - mediumPtr_dif.get(mediumPos);
#else
		const uint  estimatedPtr_medium = appEstimate(ms,mediumPos * P % L,al);
#endif
		const uint  estimatedPtr_tiny   = appEstimate(ss,tinyPos * M % P,al);
		
		if (bsum > 32){
			printf("error pos:%d bsum:%d\n",pos,bsum);
			exit(1);
		}
		const uint width = widthLg[bsum];
		const uint skip = vcOffset[largePos];
		uint x;
		if (width == 0){
			if (bsum == M) x = (uint)(-1);
			else           x = 0;
		}
		else            x = cb.enumDecode(bsum,M,buf.getBits(lp+skip+estimatedPtr_medium + estimatedPtr_tiny,width));

		uint offset = pos % M;
		x &= (((1 << offset) - 1) << (32 - offset));
		uint val = popCount32(x);
		return ls + ms + ss + val;

		//printf("pos:%d setPos:%d width:%d sm:%d im:%d ss:%d is:%d bsum:%d x:%d vcOffset:%d\n",pos,lp+estimatedPtr_medium + estimatedPtr_tiny,width,ms,mediumPos * P % L,ss,tinyPos * M % P,bsum,buf.getBits(lp+skip+estimatedPtr_medium + estimatedPtr_tiny,width),skip);
		//printf("pos:%d lp:%d ls:%d mp:%d ms:%d sp:%d ss:%d bsum:%d width:%d x:%d\n",pos,lp,ls,estimatedPtr_medium,ms,estimatedPtr_tiny,ss,bsum,width,x);		
	}

	void getBlockAndRank(uint pos, uint& previousRank, uint& x){ //32bit fix. return previousRank and x.
		const uint largePos   = pos / L;
		const uint mediumPos  = pos / P;
		const uint tinyPos    = pos / M;

		const uint lp = largePtr[largePos];
		const uint ls = largeRank[largePos];
		const uint ms = mediumRank[mediumPos];
		uint ss, bsum;
		getTinyRank(pos,ss,bsum);
		
		//const uint  estimatedPtr_medium = estimate(ms,mediumPos * P % L);
		//const uint  estimatedPtr_tiny   = estimate(ss,tinyPos * M % P);

#ifdef EN_DIF
		const uint  estimatedPtr_medium = appEstimate(ms,mediumPos * P % L,al) - mediumPtr_dif.get(mediumPos);
#else
		const uint  estimatedPtr_medium = appEstimate(ms,mediumPos * P % L,al);
#endif

		const uint  estimatedPtr_tiny   = appEstimate(ss,tinyPos * M % P,al);
		if (bsum > 32){
			printf("error pos:%d bsum:%d\n",pos,bsum);
			exit(1);
		}
		const uint width = widthLg[bsum];
		const uint skip  = vcOffset[largePos];

		previousRank = ls + ms + ss;
		if (width == 0) {
			if (bsum == M) x = (uint)(-1);
			else           x = 0;
		}
		else            x = cb.enumDecode(bsum,M,buf.getBits(lp+skip+estimatedPtr_medium + estimatedPtr_tiny,width));
	}
	
	uint getSize() const {
		return n;
	}

	uint getAllocSize()const{
		#ifdef PACK
		uint mediumRankSize = mediumRank.getSize();
		uint largePtrSize   = largePtr.getSize();
		uint largeRankSize  = largeRank.getSize();
		#else
		const uint blockL = (n + L - 1)/L;
		const uint blockM = (n + M - 1)/M; //number of tiny blocks
		const uint blockP = (n + P - 1)/P;
		uint mediumRankSize = sizeof(ushort)*blockP*8;
		uint largePtrSize = sizeof(uint)*blockL*8;
		uint largeRankSize = sizeof(uint)*blockL*8;
		#endif

		return buf.getSize()+largeRankSize+largePtrSize+mediumRankSize+mmsbV.getSize();
	}


private:
	uint n;
	uint setCount;
	uint optSize;
	uint realBufSize;
	uint vcSize;
	uint sum;
	comb cb;

	bitVector buf;	//compressed data

	uint blockL;
	uint blockM;
	uint blockP;

	

#ifdef PACK
	packedBV mediumRank;
	packedBV largePtr;
	packedBV largeRank;
#else
	ushort* mediumRank;
	uint* largePtr;
	uint* largeRank;
#endif
	
	packedBV mediumPtr_dif;
	packedBV mmsbV;
	packedBV tinyRank;
	packedBV vcOffset;
};

class compBVselect{

private:
	uchar bvElem(uchar b) const{
		if (bit) return b;
		else     return ~b;
	}

public:
	compBVselect(bitVector& bv, compBV& cbv, bool _bit): rankA(cbv), bit(_bit) {		//---: count R and Q size
		/*
		   Q is defined for all set bit.
		   Q[i] is set if it is the first set bit in a block.

		   P is defined for all blocks.
		   P[i] is set if some bit is set in a (i+1)st block
           (P is not stored explicitely)

		   R is defined for all set bit in P.
		   R[i] is set if (i)-th 1 bit of P is adjacent to the (i-1)st 1bit.
		*/
		const uint n = bv.getSize();
		rsize = 0;
		qsize = 0;
		clumpArraySize = 1;
		bool prevP = true;

		for (uint i = 0; i < n; i+= S){
			uint width = S;
			if (n-i < S) width = n-i;
			uint x = bv.getBlock(i/S);
			if (x){
				if (!prevP) clumpArraySize++;
				rsize++;
				prevP = true;
			}
			else {
				//all zero in block
				prevP = false;	
			}
			qsize += popCount32(x);
		}
		if (!prevP) clumpArraySize++;

		_rankR.resize(rsize);
		_rankQ.resize(qsize);
		clumpArray = new uint[clumpArraySize];
		if (clumpArray == NULL){
			throw "cannot allocate for _rankR and _rankQ";
		}

		prevP = true;
		uint rcount = 0;
		uint qcount = 0;
		uint clumpArrayCount = 1;
		uint cumClump = 0;
		clumpArray[0] = 0;
		for (unsigned int i = 0; i < n; i+=S){
			uint width = S;
			if (n-i < S) width = n-i;
			uint x = bv.getBlock(i/S);
			if (x){
				if (!prevP){
					_rankR.setBit(rcount++,1);
					clumpArray[clumpArrayCount++] = cumClump;
				}
				else {
					_rankR.setBit(rcount++,0);
				}

				uint pc = popCount32(x);
				_rankQ.setBit(qcount++,1);
				for (uint j = 1; j < pc; j++){
					_rankQ.setBit(qcount++,0);
				}
				prevP = true;
			}
			else {
				cumClump++;
				prevP = false;
			}
		}
		if (!prevP) clumpArray[clumpArrayCount++] = cumClump;

#ifdef SELECTNORMAL
		_rankR.build();
		_rankQ.build();
#else
		rankR.build(_rankR);
		rankQ.build(_rankQ);
#endif

		//printf("B:\n"); bv.printBits(); printf("\n");
		//printf("R:\n"); R.printBits(); printf("\n");
		//printf("Q:\n"); Q.printBits(); printf("\n");
		//printf("Clump Array:\n");
		//for (uint i = 0; i < clumpArraySize; i++){
		//	printf("%d ",clumpArray[i]);
		//}
		//printf("\n");
	}

	~compBVselect(){
		if (clumpArray) delete[] clumpArray;
	}

private:

#ifdef SELECTNORMAL
	bitVector _rankR;
	bitVector _rankQ;
#else
	compBV rankR;
	compBV rankQ;
	bitVector _rankR;
	bitVector _rankQ;
#endif
	bool bit;

	uint rsize;
	uint qsize;

	uint* clumpArray;
	uint  clumpArraySize;

	compBV& rankA;

public:
	uint select(uint num) const{
//		if (num >= rankQ.getSize()){
//			throw "error";
//		}
#ifdef SELECTNORMAL
		const uint blockRank = _rankQ.rank(num+1);
		const uint clumpNum  = _rankR.rank(blockRank);
#else
		const uint blockRank = rankQ.rank(num+1);
		const uint clumpNum  = rankR.rank(blockRank);
#endif

		const uint selectP   = clumpArray[clumpNum] + blockRank-1;
		uint previousRank = 0; //rankA.rank(selectP*S)
		uint x            = 0; //bv.getBits(selectP*S,S);
		rankA.getBlockAndRank(selectP*S,previousRank,x);
		uint remain = num + 1 - previousRank;

		//select(remain) in the block 
		uint retVal = 0;
		for (uint i = 0; i < S/8; i++){ //8 is the size of popCount
			uchar c = (x >> (S/8-i - 1)*8) & 0xFF;
			uint tmp = popCount[c];
			if (tmp >= remain) {
				retVal = selectP * S + i * 8 + selectPos[((remain-1) << 8) + c];
				break;
			}
			//else (remain > tmp)
			remain -= tmp;
		}
		return retVal;
	}

	uint getSetSize()const{
		return qsize;
	}

	uint getAllocSize() const{
#ifdef SELECTNORMAL
		const uint rankSize = _rankR.getAllocSize() + _rankQ.getAllocSize();
#else
		const uint rankSize = rankR.getAllocSize() + rankQ.getAllocSize();
#endif
		return sizeof(uint)*clumpArraySize + rankSize; //rankA is shared with rank
	}
};
