#ifndef __RR_HPP__
#define __RR_HPP__

#include "bitVector.hpp"
#include "ssv.hpp"
#include <math.h>

class rr_t{
public:
	rr_t(const bitVector& _bv): z(NULL),m(NULL),rec(0){
		size = _bv.getSize();
		const uint oneNum = _bv.getOneNum();
		float p = (float)oneNum/size;
		uint n = size;
		uint _rec = 0;
		for (float _p = p; _p < 0.5f; _p *= 2.f){
			_rec++;
		}

		z = new ssv [_rec];
		m = new uint [_rec];

		bitVector bv(size);
		bv.copy(_bv,0,size);

		uint step = 0;
		for ( ;p * 2.f < 0.5f; step++){
			uint optM = (uint)(-1.f / log(1.f - p) * log(2.f));
			uint t = 1;
			uint shift = 0;
			while (t <= optM){
				t *= 2;
				shift++;
			}
			shift--;
			m[step] = shift;
			//printf("\nstep:%d p:%f n:%d optM:%d shift:%d",step,p,n,optM,shift);
			uint curM = (1U << m[step]);
			uint nextN = 0;
			z[step].resize((n+curM-1)/curM);

			uint zeroNum = 0;
			for (uint i = 0; i < n;){
				bool allzero = true;
				for (uint t = 0; i+t < n && t < curM; t++){
					if (bv.getBit(i+t)){
						allzero = false;
						break;
					}
				}

				if (allzero){
					zeroNum++;
					z[step].setBit(i>>shift,0);
				}
				else {
					z[step].setBit(i>>shift,1);
					//copy for next step
					for (uint t = 0; i+t < n && t < curM; t++){
						bv.setBit(nextN++,bv.getBit(i+t));
					}
				}
				i += curM;
			}
			z[step].build();
			n = nextN;
			p = (float)oneNum/nextN;
		}
		//printf("\nstep:%d p:%f n:%d\n",step,p,n);
		B.resize(n);
		for (uint i = 0; i < n; i++){
			B.setBit(i,bv.getBit(i));
		}
		B.build();
		rec = step;
	}

	~rr_t() {
		delete[] z;
		delete[] m;
	}

	uint getBit(const uint pos) const {
		uint p = pos;
		for (uint i = 0; i < rec; i++){
			uint blockPos    = p >> m[i];
			uint blockOffset = p & ((1U << m[i]) - 1);
			uint b = z[i].getBit(blockPos);
			if (b == 0) return 0;
			p = ((z[i].rank(blockPos,1)-1) << m[i]) + blockOffset;
		}
		return B.getBit(p);
	}

	uint getAllocSize() const {
		uint size = B.getAllocSize();
		for (uint i = 0; i < rec; i++){
			size += z[i].getAllocSize();
		}
		size += 32*rec; //m
		return size;
	}

	uint rank(uint pos, uint bit) const{
		if (bit) return _rank1(pos);
		else     return _rank0(pos);
	}

	uint select(uint i, uint bit) const{
		if (bit) return _select1(i);
		else     return _select0(i);
	}



private:
	uint _rank1(const uint pos) const{
		uint p = pos;
		for (uint i = 0; i < rec; i++){
			uint blockPos    = p >> m[i];
			uint b = z[i].getBit(blockPos);
			uint r = z[i].rank(blockPos,1);

			if (b == 0) {              
				p = r << m[i];
				if (p == 0) return 0;
				else        p--;
			}
			else {
				uint blockOffset = p & ((1U << m[i]) - 1);
				p = ((r-1) << m[i]) + blockOffset;
			}
		}
		return B.rank(p,1);
	}

	uint _rank0(const uint pos) const{
		return pos-_rank1(pos)+1;
	}

	uint _select1(uint pos) const{
		uint p       = B.select(pos,1);
		for (uint i = rec; i > 0; ){
			i--;
			uint blockPos = p >> m[i];
			uint nextP = z[i].select(blockPos+1,1);
			uint zeroBlock = (nextP-blockPos) << m[i];
			p += zeroBlock;
		}
		return p;
	}

	uint _select0(uint pos) const{
		//binary search using rank0
		uint left  = 0;
		uint right = size;
		while (left < right){
			uint mid = (left+right)/2;
			if (_rank0(mid) < pos) left  = mid+1;
			else                   right = mid;
		}
		return left;
	}

	ssv B;
	ssv* z;
	uint* m;
	uint rec;
	uint size;
};

#endif
