#include <sdsl/vectors.hpp>

#include <iostream>
#include <ctime>
#include <sys/timeb.h>
#include <math.h>
#include "bitVector.hpp"
//#include "est.hpp"
#include "rr.hpp"
#include "vc.hpp"
#include "ssv.hpp"
//#define N (100000) //original bit vector size
#define N 1048576
#define R (10)        //original ratio of ones

//#define CHECK
#define TIME


typedef struct timeb mytimestruct;
void mygettime(mytimestruct *t)
{
  ftime(t);
}
double mylaptime(mytimestruct *before,mytimestruct *after)
{
  double t;
  t = after->time - before->time;
  t += (double)(after->millitm - before->millitm)/1000;
  return t;
}

int main(int argc, char* argv[]){
/*  double t;
  mytimestruct before,after;
  uint hoge,i,m;

  uint r = R;	// the ratio (percentage) of ones
  uint n = N;  // the length of bit vector

  srand(2); // need to be changed if necessarily

  //r = 2;
  r = 1;
  if (argc == 3){
    r = atoi(argv[1]);
    n = atoi(argv[2]);
  }

  bitVector bv(n);
  ssv sv(n); // succinct bit vector of 1 + 32/256 + 8/64 = 1.25 bpc.
  uint oneCount = 0;
  for (uint i = 0; i < n; i++){
    if ((rand() % 100) < r){
      bv.setBit(i,1);
      sv.setBit(i,1);
      oneCount++;
    }
  }
  */
//	int_vector<1> bv;
	uint hoge,i,m;
	uint r = R;
	uint n = N;
	FILE* infp = NULL;
	
	if (argc == 2){
		infp = fopen(argv[1],"rb");
		if (infp == NULL){
			printf("cannot open %s\n",argv[1]);
			return -1;
		}
		fseek(infp,0,SEEK_END);
		n = ftell(infp);
		rewind(infp);
	}
	else if (argc == 3){
		r = atoi(argv[1]);
		n = atoi(argv[2]);
	}

	bitVector bv(n);
	ssv sv(n); // succinct bit vector of 1 + 32/256 + 8/64 = 1.25 bpc.
	uint oneCount = 0;
	srand(2);
	if (!infp){
		for (uint i = 0; i < n; i++){
			if (rand() % 100 < r){
				bv.setBit(i,1);
				sv.setBit(i,1);
				oneCount++;
			}
		}
	}
	else {
		for (uint i = 0; i < n; i++){
			int c = fgetc(infp);
			if (c == EOF){
				printf("unexpected error at reading from %s\n",argv[1]);
				return -1;
			}
			if (c == '1'){
				bv.setBit(i,1);
				sv.setBit(i,1);
				oneCount++;
			}
			else if (c != '0'){
				printf("unexpected error (2) at reading from %s\n",argv[1]);
				return -1;
			}
		}
	}

  sv.build();
	
//	est_t est(bv);
  rr_t  rr(bv);
  vc_t  vc(bv);
  
  float p = (float)oneCount/n;
  printf("      h0: %11.2f\n",(oneCount * log((float)n/oneCount) + (n-oneCount) * log((float)n/(n-oneCount)))/log(2.f));
  printf("ssv size: %8d (%.2f)\n",sv.getAllocSize(),(float)sv.getAllocSize()*100.f/n);
  //	printf("est size: %8d\n",est.getAllocSize());
  printf(" rr size: %8d (%.2f)\n",rr.getAllocSize(),(float)rr.getAllocSize()*100.f/n);
  printf(" vc size: %8d\n",vc.getAllocSize());

#ifdef CHECK
	{
		//--- correct check ---
		uint rank_bv = 0;
		for (uint i = 0; i < n; i++){
			uint b = bv.getBit(i);
			if (b) rank_bv++;
			if (b){
				if (i != sv.select(rank_bv,1)) printf("error select1 rank_bv:%d bv:%d sv:%d\n",rank_bv,i,sv.select(rank_bv,1));
				//if (i != est.select(rank_bv,1)) printf("error select1 rank_bv:%d bv:%d est:%d\n",rank_bv,i,,est.select(rank_bv,1));
				if (i != rr.select(rank_bv,1))  printf("error select1 rank_bv:%d bv:%d rr:%d\n",rank_bv,i,rr.select(rank_bv,1));
//				if (i != vc.select(rank_bv,1))  printf("error select1 rank_bv:%d bv:%d vc:%d\n",rank_bv,i,vc.select(rank_bv,1));
			}
			else {
				const uint rank0_bv = i-rank_bv+1;
				if (i != sv.select(rank0_bv,0)) printf("error select0 rank_bv:%d bv:%d sv:%d\n",rank0_bv,i,sv.select(rank0_bv,0));
				//if (i != est.select(rank0_bv,0)) printf("error select0 rank_bv:%d bv:%d est:%d\n",rank0_bv,i,est.select(rank0_bv,0));
				if (i != rr.select(rank0_bv,0))  printf("error select0 rank_bv:%d bv:%d rr:%d\n",rank0_bv,i,rr.select(rank0_bv,0));
//				if (i != vc.select(rank0_bv,0))  printf("error select0 rank_bv:%d bv:%d vc:%d\n",rank0_bv,i,vc.select(rank0_bv,0));
			}
			

			if (b != sv.getBit(i)) printf("error bit pos:%d bv:%d sv:%d\n",i,bv.getBit(i),sv.getBit(i));
			//if (b != est.getBit(i)) printf("error bit pos:%d bv:%d est:%d\n",i,bv.getBit(i),est.getBit(i));
			if (b != rr.getBit(i))  printf("error bit pos:%d bv:%d rr:%d\n",i,bv.getBit(i),rr.getBit(i));
			//if (b != vc.getBit(i))  printf("error bit pos:%d bv:%d vc:%d\n",i,bv.getBit(i),vc.getBit(i));
			
			if (rank_bv != sv.rank(i,1)) printf("error rank1 pos:%d bv:%d sv:%d\n",i,rank_bv,sv.rank(i,1));
			//if (rank_bv != est.rank(i,1)) printf("error rank1 pos:%d bv:%d est:%d\n",i,rank_bv,est.rank(i,1));
			if (rank_bv != rr.rank(i,1))  printf("error rank1 pos:%d bv:%d rr:%d\n",i,rank_bv,rr.rank(i,1));
//			if (rank_bv != vc.rank(i,1))  printf("error rank1 pos:%d bv:%d vc:%d\n",i,rank_bv,vc.rank(i,1));

			// Omit the check of rank0 since it is (i-rank1)
		}
	}

	printf("CHECK PASS\n");
#endif //CHECK	

#ifdef TIME
/*
  m = oneCount;

  mygettime(&before);
  hoge = rand();
  for (i = 0; i < 100000000; i++) {
    int j;
    if (i % 10000 == 0) printf("%d\r",i);
    j = (hoge % m) + 1;
    //j = (i % m) + 1;
    sv.select(j,1);
    //rr.select(j,1);
    //vc.select(j,1);
    hoge = hoge * 1566083941U + 1;
  }
  mygettime(&after);
  t = mylaptime(&before,&after);
  printf("%d %f ssv select time\n",r,t);
  //printf("%d %f rr select time\n",r,t);
  //printf("%d %f vc select time\n",r,t);


  mygettime(&before);
  hoge = rand();
  for (i = 0; i < 100000000; i++) {
    int j;
    if (i % 10000 == 0) printf("%d\r",i);
    j = hoge % n;
    sv.rank(j,1);
    //rr.rank(j,1);
    //vc.rank(j,1);
    hoge = hoge * 1566083941U + 1;
  }
  mygettime(&after);
  t = mylaptime(&before,&after);
  printf("%d %f ssv rank time\n",r,t);
  //printf("%d %f rr rank time\n",r,t);
  //printf("%d %f vc rank time\n",r,t);
*/
#endif
//#ifdef TIME
#if 0
	uint dummy = 0;
	{
		//rank test
	  uint hoge = rand();
	  clock_t start = clock();
	  
	  for (uint i = 0; i < 10000000; i++){
			/*
			if (bv.getBit(i) != cbv.getBit(i)){
				printf("error bit  pos:%d bv:%d cbv:%d blockPos:%d \n",i,bv.getBit(i),cbv.getBit(i),i/32);
			}
			*/
			//printf("%d %d %d\n",i,bv.getBit(i),cbv.getBit(i));
			
			dummy += rr.rank(hoge%n,1);
			hoge = hoge * 1566083941U + 1;
		}
		printf("%f sec per 10^7 rank\n",((float)clock()-start)/CLOCKS_PER_SEC);
	}
	{
		//select test
	  uint hoge = rand();
	  clock_t start = clock();
	  
	  for (uint i = 0; i < 10000000; i++){
			dummy += rr.select(hoge%oneCount,1);
			hoge = hoge * 1566083941U + 1;
		}
		printf("%f sec per 10^7 select\n",((float)clock()-start)/CLOCKS_PER_SEC);
		
	}
	if (dummy == 191837311) printf("hoge\n");
#endif //TIME
{
	sdsl::stop_watch* stoppie = new sdsl::stop_watch();
	unsigned int POSITIONS = 10000000;
	unsigned int REPEATS = 10;
	unsigned int setCount = oneCount;
	unsigned int * const position = (unsigned int *)calloc( POSITIONS, sizeof *position );

	long long i;
	for(long long i = 0; i < POSITIONS; i++) position[ i ] = rand() % n;
	unsigned int dummy = 0x12345678;
	stoppie->start();	
	int k;
	for(k = REPEATS; k-- != 0; )
		for(i = 0; i < POSITIONS; i += 10 ) {
			dummy += sv.rank(position[i],1);
			dummy += sv.rank(position[i+1],1);
			dummy += sv.rank(position[i+2],1);
			dummy += sv.rank(position[i+3],1);
			dummy += sv.rank(position[i+4],1);
			dummy += sv.rank(position[i+5],1);
			dummy += sv.rank(position[i+6],1);
			dummy += sv.rank(position[i+7],1);
			dummy += sv.rank(position[i+8],1);
			dummy += sv.rank(position[i+9],1);
		}

	stoppie->stop();
	double s = stoppie->get_user_time() / 1000;
	printf( "%fs, %f ranks/s, hrg%fns/rank, r3j%f \n", s, (double)(REPEATS * POSITIONS) / s, ((double)(1E9 * s)) / (double)(REPEATS * POSITIONS), (double)sv.getAllocSize() * 8 / n );fflush(stdout);
	printf( "%d\n", dummy);
/*	float s = (float)(end - start) / CLOCKS_PER_SEC;
	printf( "%f s, %f ranks/s, %f ns/rank\n", s, (float)(REPEATS * POSITIONS) / s, ((float)(1E9 * s)) / (float)(REPEATS * POSITIONS) );fflush(stdout);
	printf("\n%d\n", dummy);*/
	for(long long i = 0; i < POSITIONS; i++) position[ i ] = rand() % setCount;
	dummy = 0x12345678;
	stoppie->start();
	
	for(k = REPEATS; k-- != 0; )
		for(i = 0; i < POSITIONS; i += 10 ) {
			dummy += sv.select(position[i],1);
			dummy += sv.select(position[i+1],1);
			dummy += sv.select(position[i+2],1);
			dummy += sv.select(position[i+3],1);
			dummy += sv.select(position[i+4],1);
			dummy += sv.select(position[i+5],1);
			dummy += sv.select(position[i+6],1);
			dummy += sv.select(position[i+7],1);
			dummy += sv.select(position[i+8],1);
			dummy += sv.select(position[i+9],1);
		}

	stoppie->stop();
	s = stoppie->get_user_time() / 1000;
	printf( "%fs, %f selects/s, hsg%fns/select, s3j%f \n", s, (float)(REPEATS * POSITIONS) / s, ((float)(1E9 * s)) / (float)(REPEATS * POSITIONS) );
	if (dummy == 42) printf("22");
}
	if (infp) fclose(infp);
	return 0;
}
