#include <iostream>
#include <ctime>
#include "bitVector.hpp"
#include "compBV.hpp"
#define N (10485760) //original bit vector size

#define CHECK
#define TIME

int main(int argc, char* argv[]){
	int r = 2;
	int n = N;
	FILE* infp = NULL;
	
	if (argc == 2){
		infp = fopen(argv[1],"rb");
		if (infp == NULL){
			printf("cannot open %s\n",argv[1]);
			return -1;
		}
		fseek(infp,0,SEEK_END);
		n = ftell(infp);
		rewind(infp);
	}
	else if (argc == 3){
		r = atoi(argv[1]);
		n = atoi(argv[2]);
	}

	bitVector bv(n);
	uint setCount = 0;
	srand(2);
	if (!infp){
		for (uint i = 0; i < n; i++){
			if (rand() % 100 < r){
				bv.setBit(i,1);
				setCount++;
			}
		}
	}
	else {
		for (uint i = 0; i < n; i++){
			int c = fgetc(infp);
			if (c == EOF){
				printf("unexpected error at reading from %s\n",argv[1]);
				return -1;
			}
			if (c == '1'){
				bv.setBit(i,1);
				setCount++;
			}
			else if (c != '0'){
				printf("unexpected error (2) at reading from %s\n",argv[1]);
				return -1;
			}
		}
	}
	/*
	for (uint i = 0; i < N/3; i++){
		if (rand() % 100 < r){
			bv.setBit(i,1);
			setCount++;
		}
	}

	for (uint i = N/3; i < N*2/3; i++){
		if (rand() % 100 < r*3){
			bv.setBit(i,1);
			setCount++;
		}
	}

	for (uint i = N*2/ 3; i < N; i++){
		if (rand() % 100 < r*4){
			bv.setBit(i,1);
			setCount++;
		}
	}
	*/

	{
		//---: gap coding size -----------
		float sum = 0.f;
		float prev = 0;

		float sum_n4 = 0.f; //nibble 4 code
		for (uint i = 0; i < n; i++){
			if (bv.getBit(i)){
				float gap = (float)i - prev + 1;
				sum += ceil(log(gap));
				sum_n4 += ceil( ceil( log(gap) / log(2.f) ) / 4.f) + 4 * floor( ceil( log(gap)/log(2.f) + 3.f) / 4.f);
				prev = (float)i;
			}
		}
		sum /= log(2.f);
		printf("gap encoding (optimal):%.2f\n",sum);
		printf("nib4 encoding         :%.2f\n",sum_n4);
	}

	{
		printf("build rank ------------\n");
		compBV cbv(bv);
		FILE* outfp = fopen("test.out","wb");
		if (outfp == NULL){
			fprintf(stderr,"cannot open test.out\n");
			return -1;
		}
		cbv.write(outfp);
		fclose(outfp);

		cbv.printAllocStatus();
	}
	{
		
		FILE* infp = fopen("test.out","rb");
		if (infp == NULL){
			fprintf(stderr,"cannot open test.out\n");
			return -1;
		}
		compBV cbv(infp);
		//cbv.printAllocStatus();
		printf("build select ------------\n");
		compBVselect cbvs(bv,cbv,true);

		printf("rank:%d select:%d \n",cbv.getAllocSize(),cbvs.getAllocSize());
		fclose(infp);

#ifdef CHECK
		{
			//correct check
			uint rank_bv = 0;
			clock_t start = clock();
			for (uint i = 0; i < n; i++){
				if (bv.getBit(i) != cbv.getBit(i)){
					printf("error bit  pos:%d bv:%d cbv:%d blockPos:%d \n",i,bv.getBit(i),cbv.getBit(i),i/32);
				}
				//printf("%d %d %d\n",i,bv.getBit(i),cbv.getBit(i));
				if (rank_bv != cbv.rank(i)){
					printf("error rank pos:%d bv:%d cbv:%d\n",i,rank_bv,cbv.rank(i));
				}
				if (bv.getBit(i)){
					if (cbvs.select(rank_bv) != i){
						printf("error select num:%d bv:%d cbvs:%d dif:%d\n",rank_bv,i,cbvs.select(rank_bv),i-cbvs.select(rank_bv));
					}
					rank_bv++;
				}
			}
		}
#endif //CHECK	
#ifdef TIME
		uint dummy = 0;
		{
			//rank test
		  uint hoge = rand();
		  clock_t start = clock();
		  
		  for (uint i = 0; i < 10000000; i++){
				/*
				if (bv.getBit(i) != cbv.getBit(i)){
					printf("error bit  pos:%d bv:%d cbv:%d blockPos:%d \n",i,bv.getBit(i),cbv.getBit(i),i/32);
				}
				*/
				//printf("%d %d %d\n",i,bv.getBit(i),cbv.getBit(i));
				
				dummy += cbv.rank(hoge%n);
				hoge = hoge * 1566083941U + 1;
			}
			printf("%f sec per 10^7 rank\n",((float)clock()-start)/CLOCKS_PER_SEC);
		}
		{
			//select test
		  uint hoge = rand();
		  clock_t start = clock();
		  
		  for (uint i = 0; i < 10000000; i++){
				dummy += cbvs.select(hoge%setCount);
				hoge = hoge * 1566083941U + 1;
			}
			printf("%f sec per 10^7 select\n",((float)clock()-start)/CLOCKS_PER_SEC);
			
		}
		if (dummy == 191837311) printf("hoge\n");
#endif //TIME
	}

	if (infp) fclose(infp);
	return 0;
}
