#ifndef __BITVECTOR_HPP__
#define __BITVECTOR_HPP__

#include "common.hpp"
using namespace std;

#define BLOCKSIZE (32)

class bitVector{
public:
	bitVector(uint _size = 0){
		size = _size;
		blockSize = (_size + BLOCKSIZE - 1)/BLOCKSIZE;
		oneNum = 0;
		B = NULL;
		build();
	}

	void resize(uint _size){	//destruct
		if (B) delete[] B;
		size = _size;
		blockSize = (_size + BLOCKSIZE - 1)/BLOCKSIZE;
		oneNum = 0;
		build();
	}

	void build(){
		if (blockSize == 0) return;
		
		B = new uint [blockSize];
		if (B == NULL){
			printf("cannoallocate for B\n");
			throw;
		}
		//memset(B,0,sizeof(uint)*blockSize);
		for (int i=0; i<blockSize; i++) B[i] = 0;
	}

	~bitVector(){
		if (B) delete[] B;
	}

	void setBit(uint pos, uint x){
		uint blockPos = pos / BLOCKSIZE;
		uint offset   = pos % BLOCKSIZE;
		if      (x == 0) B[blockPos] &= (~(1U << offset));
		else if (x == 1) B[blockPos] |=   (1U << offset);
		else {
			fprintf(stderr,"unexpected value in setBit. x = %d\n",x);
			throw;
		}
		oneNum++;
	}

	void setBits(uint pos, uint width, uint x){
		for (uint i = 0; i < width; i++){
			setBit(i+pos,(x >> i)&1);
		}
	}

	uint getBit(uint pos)const {
		uint blockPos = pos / BLOCKSIZE;
		uint offset   = pos % BLOCKSIZE;

		if (pos >= size){
			fprintf(stderr,"error in bitVector::getBit() pos:%d size:%d\n",pos,size);
			exit(1);
		}
		return (B[blockPos] >> offset) & 1;
	}

	uint getBits(uint pos, uint width) const{
		if (width == 0) return 0;
		if (width >= 32){
			fprintf(stderr,"getBits error. width:%d\n",width);
			exit(1);
		}
		if (pos+width > size){
			fprintf(stderr,"getBits error. pos:%d width:%d size:%d\n",pos,width,size);
			exit(1);
		}
		
		
		//block version
		const uint endPos = pos+width-1;
		const uint blockPos_1 = pos / BLOCKSIZE;
		const uint blockPos_2 = endPos / BLOCKSIZE;
		const uint offset_1   = pos % BLOCKSIZE;
	
		if (blockPos_1 == blockPos_2){
			return (B[blockPos_1] >> offset_1) & ((1U << width) - 1);
		}
		else {
			const uint offset_2   = endPos % BLOCKSIZE;
			return  (B[blockPos_1] >> offset_1) + ((B[blockPos_2] & ((1U << (offset_2+1))-1)) << (32-offset_1));
		}		

	}

	uint* getB() const { return B; }
	uint getBlock(const uint i) const { return B[i]; }
	uint getSize() const{ return size; }
	uint getOneNum() const { return oneNum; }
	uint getBlockSize() const{ return blockSize; }
	void printBits() const{
		for (uint i = 0; i < size; i++){
			if (getBit(i)) putchar('1');
			else           putchar('0');
			if ((i+1) % 8 == 0) putchar(' ');
			if ((i+1) % 32 == 0) putchar('\n');
		}
	}

	void copy(const bitVector& from, uint bitPos, uint bitNum){
		for (uint i = 0; i < bitNum; i++){
			uint x = from.getBit(i);
			setBit(bitPos+i,x);
		}
	}
	
private:
	uint* B;
	uint size;
	uint oneNum;
	uint blockSize;
};

#endif
