#include <iostream>
#include <ctime>
#include <sys/timeb.h>
#include <math.h>
#include "bitVector.hpp"
//#include "est.hpp"
#include "rr.hpp"
#include "vc.hpp"
#include "ssv.hpp"
//#define N (100000) //original bit vector size
#define N 1048576
#define R (10)        //original ratio of ones

#define CHECK
#define TIME


typedef struct timeb mytimestruct;
void mygettime(mytimestruct *t)
{
  ftime(t);
}
double mylaptime(mytimestruct *before,mytimestruct *after)
{
  double t;
  t = after->time - before->time;
  t += (double)(after->millitm - before->millitm)/1000;
  return t;
}

int main(int argc, char* argv[]){
  double t;
  mytimestruct before,after;
  uint hoge,i,m;

  uint r = R;	// the ratio (percentage) of ones
  uint n = N;  // the length of bit vector

  srand(2); // need to be changed if necessarily

  //r = 2;
  r = 1;
  if (argc == 3){
    r = atoi(argv[1]);
    n = atoi(argv[2]);
  }

  bitVector bv(n);
  ssv sv(n); // succinct bit vector of 1 + 32/256 + 8/64 = 1.25 bpc.
  uint oneCount = 0;
  for (uint i = 0; i < n; i++){
    if ((rand() % 100) < r){
      bv.setBit(i,1);
      sv.setBit(i,1);
      oneCount++;
    }
  }
  
  sv.build();
	
//	est_t est(bv);
  rr_t  rr(bv);
  vc_t  vc(bv);
  
  float p = (float)oneCount/n;
  printf("      h0: %11.2f\n",(oneCount * log((float)n/oneCount) + (n-oneCount) * log((float)n/(n-oneCount)))/log(2.f));
  printf("ssv size: %8d (%.2f)\n",sv.getAllocSize(),(float)sv.getAllocSize()*100.f/n);
  //	printf("est size: %8d\n",est.getAllocSize());
  printf(" rr size: %8d (%.2f)\n",rr.getAllocSize(),(float)rr.getAllocSize()*100.f/n);
  //printf(" vc size: %8d\n",vc.getAllocSize());

#ifdef CHECK
	{
		//--- correct check ---
		uint rank_bv = 0;
		for (uint i = 0; i < n; i++){
			uint b = bv.getBit(i);
			if (b) rank_bv++;
			if (b){
				if (i != sv.select(rank_bv,1)) printf("error select1 rank_bv:%d bv:%d sv:%d\n",rank_bv,i,sv.select(rank_bv,1));
				//if (i != est.select(rank_bv,1)) printf("error select1 rank_bv:%d bv:%d est:%d\n",rank_bv,i,,est.select(rank_bv,1));
				if (i != rr.select(rank_bv,1))  printf("error select1 rank_bv:%d bv:%d rr:%d\n",rank_bv,i,rr.select(rank_bv,1));
				//if (i != vc.select(rank_bv,1))  printf("error select1 rank_bv:%d bv:%d vc:%d\n",rank_bv,i,vc.select(rank_bv,1));
			}
			else {
				const uint rank0_bv = i-rank_bv+1;
				if (i != sv.select(rank0_bv,0)) printf("error select0 rank_bv:%d bv:%d sv:%d\n",rank0_bv,i,sv.select(rank0_bv,0));
				//if (i != est.select(rank0_bv,0)) printf("error select0 rank_bv:%d bv:%d est:%d\n",rank0_bv,i,est.select(rank0_bv,0));
				if (i != rr.select(rank0_bv,0))  printf("error select0 rank_bv:%d bv:%d rr:%d\n",rank0_bv,i,rr.select(rank0_bv,0));
				//if (i != vc.select(rank0_bv,0))  printf("error select0 rank_bv:%d bv:%d vc:%d\n",rank0_bv,i,vc.select(rank0_bv,0));
			}
			

			if (b != sv.getBit(i)) printf("error bit pos:%d bv:%d sv:%d\n",i,bv.getBit(i),sv.getBit(i));
			//if (b != est.getBit(i)) printf("error bit pos:%d bv:%d est:%d\n",i,bv.getBit(i),est.getBit(i));
			if (b != rr.getBit(i))  printf("error bit pos:%d bv:%d rr:%d\n",i,bv.getBit(i),rr.getBit(i));
			//if (b != vc.getBit(i))  printf("error bit pos:%d bv:%d vc:%d\n",i,bv.getBit(i),vc.getBit(i));
			
			if (rank_bv != sv.rank(i,1)) printf("error rank1 pos:%d bv:%d sv:%d\n",i,rank_bv,sv.rank(i,1));
			//if (rank_bv != est.rank(i,1)) printf("error rank1 pos:%d bv:%d est:%d\n",i,rank_bv,est.rank(i,1));
			if (rank_bv != rr.rank(i,1))  printf("error rank1 pos:%d bv:%d rr:%d\n",i,rank_bv,rr.rank(i,1));
			//if (rank_bv != vc.rank(i,1))  printf("error rank1 pos:%d bv:%d vc:%d\n",i,rank_bv,vc.rank(i,1));

			// Omit the check of rank0 since it is (i-rank1)
		}
	}

	printf("CHECK PASS\n");
#endif //CHECK	

#ifdef TIME

  m = oneCount;

  mygettime(&before);
  hoge = rand();
  for (i = 0; i < 100000000; i++) {
    int j;
    if (i % 10000 == 0) printf("%d\r",i);
    j = (hoge % m) + 1;
    //j = (i % m) + 1;
    sv.select(j,1);
    //rr.select(j,1);
    //vc.select(j,1);
    hoge = hoge * 1566083941U + 1;
  }
  mygettime(&after);
  t = mylaptime(&before,&after);
  printf("%d %f ssv select time\n",r,t);
  //printf("%d %f rr select time\n",r,t);
  //printf("%d %f vc select time\n",r,t);


  mygettime(&before);
  hoge = rand();
  for (i = 0; i < 100000000; i++) {
    int j;
    if (i % 10000 == 0) printf("%d\r",i);
    j = hoge % n;
    sv.rank(j,1);
    //rr.rank(j,1);
    //vc.rank(j,1);
    hoge = hoge * 1566083941U + 1;
  }
  mygettime(&after);
  t = mylaptime(&before,&after);
  printf("%d %f ssv rank time\n",r,t);
  //printf("%d %f rr rank time\n",r,t);
  //printf("%d %f vc rank time\n",r,t);

#endif

	return 0;
}
