#ifndef __VC_HPP__
#define __VC_HPP__

#include "common.hpp"
#include "packed_bv.hpp"
#define PACKSIZE (8) //fix. 

static const uchar mask_l[] = { 
	0x00, //00000000
	0x80, //10000000
	0xC0, //11000000
	0xE0, //11100000
	0xF0, //11110000
	0xF8, //11111000
	0xFC, //11111100
	0xFE, //11111110
};


class vc_t{
public:
	uint convert(const uint* buf, uchar* _code, uint num){
		uint mbuf = buf[0];
		for (uint i = 1; i < num; i++){
			if (buf[i] > mbuf) mbuf = buf[i];
		}
		uint mmsb = lg(mbuf);

		memset(_code,0,mmsb);
		for (uint i = 0; i < num; i++){
			for (uint j = 0; j < mmsb; j++){
				_code[j] |= ((buf[i]>>j) & 1) << (PACKSIZE-i-1);
			}
		}
		return mmsb;
	}

	~vc_t(){
		if (code) delete[] code;
	}

	vc_t(const bitVector& _bv){
		code = NULL;
		uint size = _bv.getSize();
		oneCount = _bv.getOneNum();
		uint prev = 0;
		uint* buf = new uint[PACKSIZE];
		const uint blockSize = oneCount/PACKSIZE + 1;
		uint* _pointer = new uint[blockSize+1];
		uint* _sample = new uint [blockSize+1];
		uchar* _code = new uchar[lg(size)*oneCount];

		oneCount = 0;
		uint codePos = 0;
		_sample[0] = 0;
		for (uint i = 0; i < size; i++){
			if (_bv.getBit(i)){
				buf[oneCount%PACKSIZE] = i-prev-1;
				oneCount++;
				if (oneCount%PACKSIZE == 0){
					uint mmsb = convert(buf,_code+codePos,PACKSIZE);
					_pointer[oneCount/PACKSIZE-1] = codePos;
					_sample[oneCount/PACKSIZE]  = i;
					codePos += mmsb;
				}
				prev = i;
			}
		}

		if (oneCount % PACKSIZE != 0){
			for (int i = oneCount%PACKSIZE; i < PACKSIZE; i++) buf[i] = 0;
			uint mmsb = convert(buf,_code+codePos,oneCount%PACKSIZE);
			_pointer[oneCount/PACKSIZE] = codePos;
			_pointer[oneCount/PACKSIZE+1] = codePos+mmsb;
			codePos += mmsb;
		}
		else {
			_pointer[oneCount/PACKSIZE+1] = codePos;
		}
		
		if (buf) delete[] buf;
		pointer.build(_pointer,blockSize+1);
		sample.build(_sample,blockSize);
		if (_sample) delete[] _sample;
		if (_pointer) delete[] _pointer;

		codeSize = codePos;
		code = new uchar[codeSize];
		memcpy(code,_code,codeSize);

		if (_code) delete[] _code;
	}

	uint getBit(uint pos) const {
	}

	uint rank(uint pos, uint bit) const {
		if (bit ) return _rank1(pos);
		else      return _rank0(pos);
	}

	uint select(uint pos, uint bit) const {
		if (bit) return _select1(pos);
		else     return _select0(pos);
	}

	uint getAllocSize() const {
		uint size = pointer.getSize();
		size += codeSize * 8;
		size += sample.getSize();
		return size;
	}

private:

	uint _select1(uint pos) const {
		uint p      = pos/PACKSIZE;
		uint offset = pos%PACKSIZE;
		uint start  = pointer[p];
		uint width  = pointer[p+1] - start;
		uint val    = sample[p] + offset;
		for (uint i = 0; i < width; i++) {
			val += (_popCount[code[start+i] & mask_l[offset]] << i);
		}
		return val;
	}

	uint _select0(uint pos) const{
		//binary search using select1
		uint left  = 0;
		uint right = oneCount;
		uint s1    = 0;
		while (left < right){
			uint mid = (left+right)/2;
			s1 = _select1(mid+1);
			if ((s1-mid+1) <= pos) left  = mid+1;
			else                   right = mid;
		}
		return pos+left-1;
	}

	uint _rank1(uint pos) const{
		uint left = 0;
		uint right = oneCount/PACKSIZE;
		while (left < right){
			unsigned int mid = (left+right)/2;
			if (sample[mid+1] <= pos) left = mid+1;
			else                      right = mid;
		}

		uint p      = right;
		uint start  = pointer[p];
		uint width  = pointer[p+1] - start;
		uint val    = sample[p];
		uint offset = 0;
		while (offset < PACKSIZE && (p*PACKSIZE+offset <= oneCount)){
			uint tmpVal = val+offset;
			for (uint i = 0; i < width; i++) {
				tmpVal += (_popCount[code[start+i] & mask_l[offset]] << i);
			}
			if (tmpVal > pos) break;
			offset++;
		}
		return p*PACKSIZE+offset-1;
	}

	uint _rank0(uint pos) const{
		return pos-_rank1(pos);
	}

	uint oneCount;
	uint codeSize;
	uchar* code;
	packedBV pointer;
	packedBV sample;
};

#endif
